import logging
import re
import string
from collections import defaultdict
from typing import Any, Dict, List, Optional, Set, Text, Tuple

import structlog
from jinja2 import Template
from pypred import Predicate
from pypred.ast import CompareOperator, Literal, NegateOperator

import rasa.core.training.story_conflict
import rasa.shared.nlu.constants
import rasa.shared.utils.cli
import rasa.shared.utils.io
from rasa.core.channels import UserMessage
from rasa.dialogue_understanding.stack.frames import PatternFlowStackFrame
from rasa.engine.language import Language
from rasa.exceptions import ValidationError
from rasa.shared.constants import (
    ASSISTANT_ID_DEFAULT_VALUE,
    ASSISTANT_ID_KEY,
    CONFIG_ADDITIONAL_LANGUAGES_KEY,
    CONFIG_LANGUAGE_KEY,
    CONFIG_MANDATORY_KEYS,
    CONFIG_PIPELINE_KEY,
    DOCS_URL_ACTIONS,
    DOCS_URL_DOMAIN,
    DOCS_URL_DOMAINS,
    DOCS_URL_FORMS,
    DOCS_URL_RESPONSES,
    REQUIRED_SLOTS_KEY,
    RESPONSE_CONDITION,
    UTTER_PREFIX,
)
from rasa.shared.core import constants
from rasa.shared.core.command_payload_reader import (
    MAX_NUMBER_OF_SLOTS,
    CommandPayloadReader,
)
from rasa.shared.core.constants import (
    KEY_ALLOW_NLU_CORRECTION,
    SLOTS,
    SlotMappingType,
)
from rasa.shared.core.domain import (
    RESPONSE_KEYS_TO_INTERPOLATE,
    Domain,
)
from rasa.shared.core.events import ActionExecuted, ActiveLoop, UserUttered
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.flows.constants import KEY_NAME, KEY_TRANSLATION
from rasa.shared.core.flows.flow_step_links import IfFlowStepLink
from rasa.shared.core.flows.steps.action import ActionFlowStep
from rasa.shared.core.flows.steps.collect import CollectInformationFlowStep
from rasa.shared.core.flows.steps.link import LinkFlowStep
from rasa.shared.core.flows.steps.set_slots import SetSlotsFlowStep
from rasa.shared.core.flows.utils import (
    get_duplicate_slot_persistence_config_error_message,
    get_invalid_slot_persistence_config_error_message,
    warn_deprecated_collect_step_config,
)
from rasa.shared.core.generator import TrainingDataGenerator
from rasa.shared.core.slot_mappings import CoexistenceSystemType
from rasa.shared.core.slots import BooleanSlot, CategoricalSlot, ListSlot, Slot
from rasa.shared.core.training_data.story_reader.yaml_story_reader import (
    YAMLStoryReader,
)
from rasa.shared.core.training_data.structures import StoryGraph
from rasa.shared.data import create_regex_pattern_reader
from rasa.shared.exceptions import RasaException
from rasa.shared.importers.importer import TrainingDataImporter
from rasa.shared.nlu.constants import COMMANDS
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.nlu.training_data.training_data import TrainingData
from rasa.telemetry import track_validation_error_log

logger = logging.getLogger(__name__)

structlog_processors = structlog.get_config()["processors"]
updated_processors = [track_validation_error_log] + structlog_processors
structlogger = structlog.get_logger(processors=updated_processors)


class Validator:
    """A class used to verify usage of intents and utterances."""

    def __init__(
        self,
        domain: Domain,
        intents: TrainingData,
        story_graph: StoryGraph,
        flows: FlowsList,
        config: Optional[Dict[Text, Any]],
    ) -> None:
        """Initializes the Validator object.

        Args:
            domain: The domain.
            intents: Training data.
            story_graph: The story graph.
            flows: The flows.
            config: The configuration.
        """
        self.domain = domain
        self.intents = intents
        self.story_graph = story_graph
        self.flows = flows
        self.config = config or {}

    @classmethod
    def from_importer(cls, importer: TrainingDataImporter) -> "Validator":
        """Create an instance from the domain, nlu and story files."""
        domain = importer.get_domain()
        story_graph = importer.get_stories()
        intents = importer.get_nlu_data()
        config = importer.get_config()
        flows = importer.get_flows()

        return cls(domain, intents, story_graph, flows, config)

    def _non_default_intents(self) -> List[Text]:
        return [
            item
            for item in self.domain.intents
            if item not in constants.DEFAULT_INTENTS
        ]

    def verify_intents(self, ignore_warnings: bool = True) -> bool:
        """Compares list of intents in domain with intents in NLU training data."""
        everything_is_alright = True

        nlu_data_intents = {e.data["intent"] for e in self.intents.intent_examples}

        for intent in self._non_default_intents():
            if intent not in nlu_data_intents:
                structlogger.warn(
                    "validator.verify_intents.not_in_nlu_training_data",
                    intent=intent,
                    event_info=(
                        f"The intent '{intent}' is listed "
                        f"in the domain file, but is not found "
                        f"in the NLU training data."
                    ),
                )
                everything_is_alright = ignore_warnings or everything_is_alright

        for intent in nlu_data_intents:
            if intent not in self.domain.intents:
                structlogger.warn(
                    "validator.verify_intents.not_in_domain",
                    intent=intent,
                    event_info=(
                        f"There is a message in the training data "
                        f"labeled with intent '{intent}'. This "
                        f"intent is not listed in your domain. You "
                        f"should need to add that intent to your domain "
                        f"file!"
                    ),
                    docs=DOCS_URL_DOMAINS,
                )
                everything_is_alright = ignore_warnings

        return everything_is_alright

    def verify_example_repetition_in_intents(
        self, ignore_warnings: bool = True
    ) -> bool:
        """Checks if there is no duplicated example in different intents."""
        everything_is_alright = True

        duplication_hash = defaultdict(set)
        for example in self.intents.intent_examples:
            text = example.get(rasa.shared.nlu.constants.TEXT)
            duplication_hash[text].add(example.get("intent"))

        for text, intents in duplication_hash.items():
            if len(duplication_hash[text]) > 1:
                everything_is_alright = ignore_warnings
                intents_string = ", ".join(sorted(intents))
                structlogger.warn(
                    "validator.verify_example_repetition_in_intents"
                    ".one_example_multiple_intents",
                    example=text,
                    intents=intents_string,
                    event_info=(
                        f"The example '{text}' was found labeled "
                        f"with multiple different intents in the "
                        f"training data. Each annotated message "
                        f"should only appear with one intent. "
                        f"You should fix that conflict The example is "
                        f"labeled with: {intents_string}."
                    ),
                )
        return everything_is_alright

    def verify_intents_in_stories_or_flows(self, ignore_warnings: bool = True) -> bool:
        """Checks intents used in stories.

        Verifies if the intents used in the stories are valid, and whether
        all valid intents are used in the stories.
        """
        everything_is_alright = self.verify_intents(ignore_warnings=ignore_warnings)

        stories_intents = {
            event.intent["name"]
            for story in self.story_graph.story_steps
            for event in story.events
            if type(event) == UserUttered and event.intent_name is not None
        }
        flow_intents = {
            trigger.intent
            for flow in self.flows.underlying_flows
            if flow.nlu_triggers is not None
            for trigger in flow.nlu_triggers.trigger_conditions
        }
        used_intents = stories_intents.union(flow_intents)

        for intent in used_intents:
            if intent not in self.domain.intents:
                structlogger.warn(
                    "validator.verify_intents_in_stories_or_flows.not_in_domain",
                    intent=intent,
                    event_info=(
                        f"The intent '{intent}' is used in a "
                        f"story or flow, but it is not listed in "
                        f"the domain file. You should add it to your "
                        f"domain file!"
                    ),
                    docs=DOCS_URL_DOMAINS,
                )
                everything_is_alright = ignore_warnings

        for intent in self._non_default_intents():
            if intent not in used_intents:
                structlogger.warn(
                    "validator.verify_intents_in_stories_or_flows.not_used",
                    intent=intent,
                    event_info=(
                        f"The intent '{intent}' is not used in any story, rule or flow."
                    ),
                )
                everything_is_alright = ignore_warnings or everything_is_alright

        return everything_is_alright

    def _gather_utterance_actions(self) -> Set[Text]:
        """Return all utterances which are actions.

        Returns:
            A set of response names found in the domain and data files, with the
            response key stripped in the case of response selector responses.
        """
        domain_responses = {
            response.split(rasa.shared.nlu.constants.RESPONSE_IDENTIFIER_DELIMITER)[0]
            for response in self.domain.responses.keys()
            if response in self.domain.action_names_or_texts
        }
        data_responses = {
            response.split(rasa.shared.nlu.constants.RESPONSE_IDENTIFIER_DELIMITER)[0]
            for response in self.intents.responses.keys()
        }
        return domain_responses.union(data_responses)

    def _does_story_only_use_valid_actions(
        self, used_utterances_in_stories: Set[str], utterance_actions: List[str]
    ) -> bool:
        """Checks if all utterances used in stories are valid."""
        has_no_warnings = True
        for used_utterance in used_utterances_in_stories:
            if used_utterance not in utterance_actions:
                structlogger.warn(
                    "validator.invalid_utterance_action",
                    action=used_utterance,
                    event_info=(
                        f"The action '{used_utterance}' is used in the stories, "
                        f"but is not a valid utterance action. Please make sure "
                        f"the action is listed in your domain and there is a "
                        f"template defined with its name."
                    ),
                    docs=DOCS_URL_ACTIONS + "#utterance-actions",
                )
                has_no_warnings = False
        return has_no_warnings

    def _utterances_used_in_stories(self) -> Set[str]:
        """Return all utterances which are used in stories."""
        stories_utterances = set()

        for story in self.story_graph.story_steps:
            for event in story.events:
                if not isinstance(event, ActionExecuted):
                    continue

                if not event.action_name:
                    continue

                if not event.action_name.startswith(UTTER_PREFIX):
                    # we are only interested in utter actions
                    continue

                if event.action_name in stories_utterances:
                    # we already processed this one before, we only want to warn once
                    continue

                stories_utterances.add(event.action_name)
        return stories_utterances

    @classmethod
    def check_for_placeholder(cls, value: Any) -> bool:
        """Check if a value contains a placeholder."""
        if isinstance(value, str):
            return bool(re.search(r"{\s*}", value))
        elif isinstance(value, dict):
            return any(cls.check_for_placeholder(i) for i in value.values())
        elif isinstance(value, list):
            return any(cls.check_for_placeholder(i) for i in value)
        return False

    def check_for_no_empty_parenthesis_in_responses(self) -> bool:
        """Checks if there are no empty parenthesis in utterances."""
        everything_is_alright = True

        for response_text, response_variations in self.domain.responses.items():
            if not response_variations:
                structlogger.error(
                    "validator.empty_response",
                    response=response_text,
                    event_info=(
                        f"The response '{response_text}' in the domain file "
                        f"does not have any variations. Please add at least one "
                        f"variation to the response."
                    ),
                )
                everything_is_alright = False

            for response in response_variations:
                if any(
                    self.check_for_placeholder(response.get(key))
                    for key in RESPONSE_KEYS_TO_INTERPOLATE
                ):
                    structlogger.error(
                        "validator.empty_parenthesis_in_utterances",
                        response=response_text,
                        event_info=(
                            f"The response '{response_text}' in the domain file "
                            f"contains empty parenthesis, which is not permitted. "
                            f"Please remove the empty parenthesis."
                        ),
                    )
                    everything_is_alright = False

        return everything_is_alright

    def verify_forms_in_stories_rules(self) -> bool:
        """Verifies that forms referenced in active_loop directives are present."""
        all_forms_exist = True
        visited_loops = set()

        for story in self.story_graph.story_steps:
            for event in story.events:
                if not isinstance(event, ActiveLoop):
                    continue

                if event.name in visited_loops:
                    # We've seen this loop before, don't alert on it twice
                    continue

                if not event.name:
                    # To support setting `active_loop` to `null`
                    continue

                if event.name not in self.domain.action_names_or_texts:
                    structlogger.error(
                        "validator.verify_forms_in_stories_rules.not_in_domain",
                        form=event.name,
                        block=story.block_name,
                        event_info=(
                            f"The form '{event.name}' is used in the "
                            f"'{story.block_name}' block, but it "
                            f"is not listed in the domain file. "
                            f"You should add it to your "
                            f"domain file!"
                        ),
                        docs=DOCS_URL_FORMS,
                    )
                    all_forms_exist = False
                visited_loops.add(event.name)

        return all_forms_exist

    def verify_actions_in_stories_rules(self) -> bool:
        """Verifies that actions used in stories and rules are present in the domain."""
        everything_is_alright = True
        visited = set()

        for story in self.story_graph.story_steps:
            for event in story.events:
                if not isinstance(event, ActionExecuted):
                    continue

                if not event.action_name:
                    continue

                if not event.action_name.startswith("action_"):
                    continue

                if event.action_name in visited:
                    # we already processed this one before, we only want to warn once
                    continue

                if event.action_name not in self.domain.action_names_or_texts:
                    structlogger.error(
                        "validator.verify_actions_in_stories_rules.not_in_domain",
                        action=event.action_name,
                        block=story.block_name,
                        event_info=(
                            f"The action '{event.action_name}' is used in the "
                            f"'{story.block_name}' block, but it "
                            f"is not listed in the domain file. You "
                            f"should add it to your domain file!"
                        ),
                        docs=DOCS_URL_DOMAINS,
                    )
                    everything_is_alright = False
                visited.add(event.action_name)

        return everything_is_alright

    def verify_story_structure(
        self, ignore_warnings: bool = True, max_history: Optional[int] = None
    ) -> bool:
        """Verifies that the bot behaviour in stories is deterministic.

        Args:
            ignore_warnings: When `True`, return `True` even if conflicts were found.
            max_history: Maximal number of events to take into account for conflict
                identification.

        Returns:
            `False` is a conflict was found and `ignore_warnings` is `False`.
            `True` otherwise.
        """
        structlogger.info(
            "validator.verify_story_structure.start",
            event_info="Story structure validation...",
        )

        trackers = TrainingDataGenerator(
            self.story_graph,
            domain=self.domain,
            remove_duplicates=False,
            augmentation_factor=0,
        ).generate_story_trackers()

        # Create a list of `StoryConflict` objects
        conflicts = rasa.core.training.story_conflict.find_story_conflicts(
            trackers, self.domain, max_history
        )

        if not conflicts:
            structlogger.info(
                "validator.verify_story_structure.no_conflicts",
                event_info="No story structure conflicts found.",
            )
        else:
            for conflict in conflicts:
                structlogger.warn(
                    "validator.verify_story_structure.conflicts",
                    event_info="Found story structure conflict",
                    conflict=str(conflict),
                )

        return ignore_warnings or not conflicts

    def verify_nlu(self, ignore_warnings: bool = True) -> bool:
        """Runs all the validations on intents and utterances."""
        structlogger.info(
            "validator.verify_intents_in_stories.start",
            event_info="Validating intents...",
        )
        intents_are_valid = self.verify_intents_in_stories_or_flows(ignore_warnings)

        structlogger.info(
            "validator.verify_example_repetition_in_intents.start",
            event_info="Validating uniqueness of intents and stories...",
        )
        there_is_no_duplication = self.verify_example_repetition_in_intents(
            ignore_warnings
        )

        return intents_are_valid and there_is_no_duplication

    def verify_form_slots(self) -> bool:
        """Verifies that form slots match the slot mappings in domain."""
        domain_slot_names = [slot.name for slot in self.domain.slots]
        everything_is_alright = True

        for form in self.domain.form_names:
            form_slots = self.domain.required_slots_for_form(form)
            for slot in form_slots:
                if slot in domain_slot_names:
                    continue
                else:
                    structlogger.warn(
                        "validator.verify_form_slots.not_in_domain",
                        slot=slot,
                        form=form,
                        event_info=(
                            f"The form slot '{slot}' in form '{form}' "
                            f"is not present in the domain slots."
                            f"Please add the correct slot or check for typos."
                        ),
                        docs=DOCS_URL_DOMAINS,
                    )
                    everything_is_alright = False

        return everything_is_alright

    def verify_slot_mappings(self) -> bool:
        """Verifies that slot mappings match forms."""
        everything_is_alright = True

        for slot in self.domain.slots:
            for mapping in slot.mappings:
                for condition in mapping.conditions:
                    condition_active_loop = condition.active_loop
                    mapping_type = mapping.type
                    if (
                        condition_active_loop
                        and condition_active_loop not in self.domain.form_names
                    ):
                        structlogger.error(
                            "validator.verify_slot_mappings.not_in_domain",
                            slot=slot.name,
                            form=condition_active_loop,
                            event_info=(
                                f"Slot '{slot.name}' has a mapping "
                                f"condition for form '{condition_active_loop}' "
                                f"which is not listed in domain forms. Please "
                                f"add this form to the forms section or check "
                                f"for typos."
                            ),
                        )
                        everything_is_alright = False

                    form_slots = self.domain.forms.get(condition_active_loop, {}).get(
                        REQUIRED_SLOTS_KEY, {}
                    )
                    if (
                        form_slots
                        and slot.name not in form_slots
                        and mapping_type != SlotMappingType.FROM_TRIGGER_INTENT
                    ):
                        structlogger.warn(
                            "validator.verify_slot_mappings.not_in_forms_key",
                            slot=slot.name,
                            form=condition_active_loop,
                            forms_key=REQUIRED_SLOTS_KEY,
                            event_info=(
                                f"Slot '{slot.name}' has a mapping condition "
                                f"for form '{condition_active_loop}', but it's "
                                f"not present in '{condition_active_loop}' "
                                f"form's '{REQUIRED_SLOTS_KEY}'. "
                                f"The slot needs to be added to this key."
                            ),
                        )

        return everything_is_alright

    def verify_domain_validity(self) -> bool:
        """Checks whether the domain returned by the importer is empty.

        An empty domain or one that uses deprecated Mapping Policy is invalid.
        """
        if self.domain.is_empty():
            return False

        for intent_key, intent_dict in self.domain.intent_properties.items():
            if "triggers" in intent_dict:
                structlogger.warn(
                    "validator.verify_domain_validity.mapping_policy_deprecation",
                    intent_key=intent_key,
                    event_info=(
                        f"The intent {intent_key} in the domain file "
                        f"is using the MappingPolicy format "
                        f"which has now been deprecated. "
                        f"Please migrate to RulePolicy."
                    ),
                )
                return False

        return True

    def warn_if_config_mandatory_keys_are_not_set(self) -> None:
        """Raises a warning if mandatory keys are not present in the config.

        Additionally, raises a UserWarning if the assistant_id key is filled with the
        default placeholder value.
        """
        for key in set(CONFIG_MANDATORY_KEYS):
            if key not in self.config:
                structlogger.warn(
                    "validator.config_missing_mandatory_key",
                    key=key,
                    event_info=(
                        f"The config file is missing the '{key}' mandatory key."
                    ),
                )

        assistant_id = self.config.get(ASSISTANT_ID_KEY)

        if assistant_id is not None and assistant_id == ASSISTANT_ID_DEFAULT_VALUE:
            structlogger.warn(
                "validator.config_missing_unique_mandatory_key_value",
                key=ASSISTANT_ID_KEY,
                event_info=(
                    f"The config file is missing a unique value for the "
                    f"'{ASSISTANT_ID_KEY}' mandatory key. Please replace the default "
                    f"placeholder value with a unique identifier."
                ),
            )

    def _log_error_if_either_action_or_utterance_are_not_defined(
        self,
        collect: CollectInformationFlowStep,
        all_good: bool,
        domain_slots: Dict[Text, Slot],
        flow_id: str,
    ) -> bool:
        """Validates that a collect step can have either an action or an utterance.
        Also logs an error if neither an action nor an utterance is defined.

        Args:
            collect: the name of the slot to collect
            all_good: boolean value indicating the validation status

        Returns:
            False, if validation failed, true, otherwise
        """
        has_utterance_defined = any(
            [u for u in self.domain.utterances_for_response if u == collect.utter]
        )

        has_action_defined = any(
            [
                a
                for a in self.domain.action_names_or_texts
                if a == collect.collect_action
            ]
        )

        if has_utterance_defined and has_action_defined:
            structlogger.error(
                "validator.verify_flows_steps_against_domain.collect_step",
                collect=collect.collect,
                has_utterance_defined=has_utterance_defined,
                has_action_defined=has_action_defined,
                flow=flow_id,
                event_info=(
                    f"The collect step '{collect.collect}' has an utterance "
                    f"'{collect.utter}' as well as an action "
                    f"'{collect.collect_action}' defined. "
                    f"You can just have one of them! "
                    f"Please remove either the utterance or the action."
                ),
            )
            all_good = False

        slot = domain_slots.get(collect.collect)
        slot_has_initial_value_defind = slot and slot.initial_value is not None

        if (
            not slot_has_initial_value_defind
            and not has_utterance_defined
            and not has_action_defined
        ):
            structlogger.error(
                "validator.verify_flows_steps_against_domain.collect_step",
                collect=collect.collect,
                has_utterance_defined=has_utterance_defined,
                has_action_defined=has_action_defined,
                flow=flow_id,
                event_info=(
                    f"The collect step '{collect.collect}' has neither an utterance "
                    f"nor an action defined, or an initial value defined in the domain."
                    f"You need to define either an utterance or an action."
                ),
            )
            all_good = False

        return all_good

    @staticmethod
    def _log_error_if_slot_not_in_domain(
        slot_name: str,
        domain_slots: Dict[Text, Slot],
        step_id: str,
        flow_id: str,
        all_good: bool,
    ) -> bool:
        if slot_name not in domain_slots:
            structlogger.error(
                "validator.verify_flows_steps_against_domain.slot_not_in_domain",
                slot=slot_name,
                step=step_id,
                flow=flow_id,
                event_info=(
                    f"The slot '{slot_name}' is used in the "
                    f"step '{step_id}' of flow id '{flow_id}', but it "
                    f"is not listed in the domain slots. "
                    f"You should add it to your domain file!"
                ),
            )
            all_good = False

        return all_good

    @staticmethod
    def _log_error_if_list_slot(
        slot: Slot, step_id: str, flow_id: str, all_good: bool
    ) -> bool:
        if isinstance(slot, ListSlot):
            structlogger.error(
                "validator.verify_flows_steps_against_domain.use_of_list_slot_in_flow",
                slot=slot.name,
                step=step_id,
                flow=flow_id,
                event_info=(
                    f"The slot '{slot.name}' is used in the "
                    f"step '{step_id}' of flow id '{flow_id}', but it "
                    f"is a list slot. List slots are currently not "
                    f"supported in flows. You should change it to a "
                    f"text, boolean or float slot in your domain file!"
                ),
            )
            all_good = False

        return all_good

    def verify_flows_steps_against_domain(self) -> bool:
        """Checks flows steps' references against the domain file."""
        all_good = True
        domain_slots = {slot.name: slot for slot in self.domain.slots}
        flow_ids = [flow.id for flow in self.flows.underlying_flows]

        for flow in self.flows.underlying_flows:
            for step in flow.steps_with_calls_resolved:
                if isinstance(step, CollectInformationFlowStep):
                    all_good = (
                        self._log_error_if_either_action_or_utterance_are_not_defined(
                            step, all_good, domain_slots, flow.id
                        )
                    )

                    all_good = self._log_error_if_slot_not_in_domain(
                        step.collect, domain_slots, step.id, flow.id, all_good
                    )

                    current_slot = domain_slots.get(step.collect)
                    if not current_slot:
                        continue

                    all_good = self._log_error_if_list_slot(
                        current_slot, step.id, flow.id, all_good
                    )

                elif isinstance(step, SetSlotsFlowStep):
                    for slot in step.slots:
                        slot_name = slot["key"]
                        all_good = self._log_error_if_slot_not_in_domain(
                            slot_name, domain_slots, step.id, flow.id, all_good
                        )
                        current_slot = domain_slots.get(slot_name)
                        if not current_slot:
                            continue

                        all_good = self._log_error_if_list_slot(
                            current_slot, step.id, flow.id, all_good
                        )

                elif isinstance(step, ActionFlowStep):
                    regex = r"{context\..+?}"
                    matches = re.findall(regex, step.action)
                    if matches:
                        structlogger.debug(
                            "validator.verify_flows_steps_against_domain"
                            ".interpolated_action",
                            action=step.action,
                            step=step.id,
                            flow=flow.id,
                            event_info=(
                                f"An interpolated action name '{step.action}' was "
                                f"found at step '{step.id}' of flow id '{flow.id}'. "
                                f"Skipping validation for this step. "
                                f"Please make sure that the action name is "
                                f"listed in your domain responses or actions."
                            ),
                        )
                    elif step.action not in self.domain.action_names_or_texts:
                        structlogger.error(
                            "validator.verify_flows_steps_against_domain"
                            ".action_not_in_domain",
                            action=step.action,
                            step=step.id,
                            flow=flow.id,
                            event_info=(
                                f"The action '{step.action}' is used in the "
                                f"step '{step.id}' of flow id '{flow.id}', "
                                f"but it is not listed in the domain file. "
                                f"You should add it to your domain file!"
                            ),
                        )
                        all_good = False

                elif isinstance(step, LinkFlowStep):
                    if step.link not in flow_ids:
                        logger.error(
                            f"The flow '{step.link}' is used in the step "
                            f"'{step.id}' of flow id '{flow.id}', but it "
                            f"is not listed in the flows file. "
                            f"Did you make a typo?",
                        )
                        all_good = False
        return all_good

    def verify_unique_flows(self) -> bool:
        """Checks if all flows have unique names and descriptions."""
        all_good = True
        flow_names = set()
        flow_descriptions = set()
        punctuation_table = str.maketrans({i: "" for i in string.punctuation})

        for flow in self.flows.underlying_flows:
            flow_description = flow.description
            cleaned_description = flow_description.translate(punctuation_table)  # type: ignore[union-attr]
            if cleaned_description in flow_descriptions:
                structlogger.error(
                    "validator.verify_unique_flows.duplicate_description",
                    flow=flow.id,
                    event_info=(
                        f"Detected duplicate flow description for "
                        f"flow id '{flow.id}'. Flow descriptions must be "
                        f"unique. Please make sure that all flows have "
                        f"different descriptions."
                    ),
                )
                all_good = False

            if flow.name in flow_names:
                structlogger.error(
                    "validator.verify_unique_flows.duplicate_name",
                    flow=flow.id,
                    name=flow.name,
                    event_info=(
                        f"Detected duplicate flow name '{flow.name}' for flow "
                        f"id '{flow.id}'. Flow names must be unique. "
                        f"Please make sure that all flows have different names."
                    ),
                )
                all_good = False

            flow_names.add(flow.name)
            flow_descriptions.add(cleaned_description)

        return all_good

    def _build_context(self) -> Dict[str, Any]:
        """Build context for jinja template rendering.

        Returns:
            A dictionary containing the allowed namespaces for jinja template rendering:
            - `context`: The context mapping the attributes of every flow stack frame
                to None values because this is only used for rendering the template
                during validation.
            - `slots`: The slots of the domain mapped to None values because this is
                only used for rendering the template during validation and not for
                evaluating the predicate at runtime.
        """
        subclasses = [subclass for subclass in PatternFlowStackFrame.__subclasses__()]
        subclass_attrs = []
        for subclass in subclasses:
            subclass_attrs.extend(
                [attr for attr in dir(subclass) if not attr.startswith("__")]
            )

        context = {
            "context": {attr: None for attr in subclass_attrs},
            "slots": {slot.name: None for slot in self.domain.slots},
        }
        return context

    @staticmethod
    def _construct_predicate(
        predicate: Optional[str],
        object_id: str,
        context: Dict[str, Any],
        is_step: bool,
        all_good: bool = True,
    ) -> Tuple[Optional[Predicate], bool]:
        rendered_template = Template(predicate).render(context)
        try:
            pred = Predicate(rendered_template)
        except (TypeError, Exception) as exception:
            if is_step:
                structlogger.error(
                    "validator.verify_predicates.step_predicate.error",
                    step=object_id,
                    exception=exception,
                    event_info=(
                        f"Could not initialize the predicate found under step "
                        f"'{object_id}': {exception}"
                    ),
                )
            elif object_id.startswith("utter_"):
                structlogger.error(
                    "validator.validate_conditional_response_variation_predicates.error",
                    utter=object_id,
                    exception=exception,
                    event_info=(
                        f"Could not initialize the predicate found under response "
                        f"variation '{object_id}': {exception}"
                    ),
                )
            else:
                structlogger.error(
                    "validator.verify_predicates.flow_guard_predicate.error",
                    flow=object_id,
                    exception=exception,
                    event_info=(
                        f"Could not initialize the predicate found in flow guard "
                        f"for flow: '{object_id}': {exception}."
                    ),
                )
            pred = None
            all_good = False

        return pred, all_good

    def _extract_predicate_syntax_tree(self, predicate: Predicate) -> Any:
        """Extract the predicate syntax tree from the given predicate.

        Args:
            predicate: The predicate from which to extract the syntax tree.

        Returns:
            The extracted syntax tree.
        """
        if isinstance(predicate.ast, NegateOperator):
            return predicate.ast.left
        return predicate.ast

    def _extract_slot_name_and_slot_value(
        self,
        predicate_syntax_tree: Any,
    ) -> Tuple[Optional[List[str]], Optional[Any]]:
        """Extract the slot name and slot value from the predicate syntax tree.

        Args:
            predicate_syntax_tree: The predicate syntax tree.

        Returns:
            A tuple containing the slot name and slot value.
        """
        try:
            if isinstance(predicate_syntax_tree.left, Literal):
                slot_name = predicate_syntax_tree.left.value.split(".")
                slot_value = predicate_syntax_tree.right.value
            else:
                slot_name = predicate_syntax_tree.right.value.split(".")
                slot_value = predicate_syntax_tree.left.value
        except AttributeError:
            # predicate only has negation and doesn't need to be checked further
            return None, None
        return slot_name, slot_value

    def _validate_categorical_value_check(
        self,
        slot_name: str,
        slot_value: Any,
        valid_slot_values: List[str],
        all_good: bool,
        step_id: str,
        link_condition: str,
        flow_id: str,
    ) -> bool:
        """Validates the categorical slot check.

        Validates that the categorical slot is checked against valid values.

        Args:
            slot_name: name of the slot to be checked
            slot_value: value of the slot to be checked
            valid_slot_values: valid values for the given slot
            all_good: flag whether all the validations have passed so far
            step_id: id of the step in which the values are being checked
            link_condition: condition where the values are being checked
            flow_id: id of the flow where the values are being checked

        Returns:
            False, if validation failed, previous value of all_good, otherwise
        """
        valid_slot_values.append(None)
        # slot_value can either be None, a string or a list of Literal objects
        if slot_value is None:
            slot_value = [None]
        if isinstance(slot_value, str):
            slot_value = [Literal(slot_value)]

        slot_values_validity = [
            sv is None
            or re.sub(r'^[\'"](.+)[\'"]$', r"\1", sv.value) in valid_slot_values
            for sv in slot_value
        ]
        if not all(slot_values_validity):
            structlogger.error(
                "validator.verify_predicates.link.invalid_condition",
                step=step_id,
                link=link_condition,
                flow=flow_id,
                event_info=(
                    f"Detected invalid condition '{link_condition}' "
                    f"at step '{step_id}' for flow id '{flow_id}'. "
                    f"The condition contains invalid values for slot {slot_name}. "
                    f"Please make sure that all conditions are valid."
                ),
            )
            return False
        return all_good

    def _validate_categorical_and_boolean_values_check(
        self,
        predicate: Predicate,
        all_good: bool,
        step_id: str,
        link_condition: str,
        flow_id: str,
    ) -> bool:
        """Validates the categorical and boolean slot checks.

        Validates that the categorical and boolean slots
        are checked against valid values.

        Args:
            predicate: condition that is supposed to be validated
            all_good: flag whether all the validations have passed so far
            step_id: id of the step in which the values are being checked
            link_condition: condition where the values are being checked
            flow_id: id of the flow where the values are being checked

        Returns:
            False, if validation failed, previous value of all_good, otherwise
        """
        predicate_syntax_tree = self._extract_predicate_syntax_tree(predicate)
        slot_namespace, slot_value = self._extract_slot_name_and_slot_value(
            predicate_syntax_tree
        )

        if slot_namespace is None:
            return all_good

        if slot_namespace[0] == "slots":
            slot_name = slot_namespace[1]
            # slots.{{context.variable}} gets evaluated to `slots.None`,
            # these predicates can only be validated during runtime
            if slot_name == "None":
                return all_good
        else:
            return all_good

        try:
            slot = next(slot for slot in self.domain.slots if slot.name == slot_name)
        except StopIteration:
            structlogger.error(
                "validator.verify_predicates.link.invalid_condition",
                step=step_id,
                link=link_condition,
                flow=flow_id,
                event_info=(
                    f"Detected invalid condition '{link_condition}' "
                    f"at step '{step_id}' for flow id '{flow_id}'. "
                    f"Slot {slot_name} is not defined in the domain file. "
                    f"Please make sure that all conditions are valid."
                ),
            )
            return False
        if isinstance(slot, CategoricalSlot):
            return self._validate_categorical_value_check(
                slot_name,
                slot_value,
                slot.values,
                all_good,
                step_id,
                link_condition,
                flow_id,
            )

        if (
            isinstance(slot, BooleanSlot)
            and isinstance(predicate_syntax_tree, CompareOperator)
            and not isinstance(predicate_syntax_tree.right.value, bool)
        ):
            structlogger.error(
                "validator.verify_predicates.link.invalid_condition",
                step=step_id,
                link=link_condition,
                flow=flow_id,
                event_info=(
                    f"Detected invalid condition '{link_condition}' "
                    f"at step '{step_id}' for flow id '{flow_id}'. "
                    f"Boolean slots can only be compared to "
                    f"boolean values (true, false). "
                    f"Please make sure that all conditions are valid."
                ),
            )
            return False
        return all_good

    def verify_predicates(self) -> bool:
        """Validate predicates used in flow step links and slot rejections."""
        all_good = True
        context = self._build_context()

        for flow in self.flows.underlying_flows:
            if flow.guard_condition:
                predicate, all_good = self._construct_predicate(
                    flow.guard_condition,
                    flow.id,
                    context,
                    is_step=False,
                    all_good=all_good,
                )
                if predicate and not predicate.is_valid():
                    structlogger.error(
                        "validator.verify_predicates.flow_guard.invalid_condition",
                        flow=flow.id,
                        flow_guard=flow.guard_condition,
                        event_info=(
                            f"Detected invalid flow guard condition "
                            f"'{flow.guard_condition}' for flow id '{flow.id}'. "
                            f"Please make sure that all conditions are valid."
                        ),
                    )
                    all_good = False
            for step in flow.steps_with_calls_resolved:
                for link in step.next.links:
                    if isinstance(link, IfFlowStepLink):
                        all_good = self._verify_namespaces(
                            link.condition, step.id, flow.id, all_good
                        )

                        predicate, all_good = self._construct_predicate(
                            link.condition,
                            step.id,
                            context,
                            is_step=True,
                            all_good=all_good,
                        )
                        if predicate and not predicate.is_valid():
                            structlogger.error(
                                "validator.verify_predicates.link.invalid_condition",
                                step=step.id,
                                link=link.condition,
                                flow=flow.id,
                                event_info=(
                                    f"Detected invalid condition '{link.condition}' "
                                    f"at step '{step.id}' for flow id '{flow.id}'. "
                                    f"Please make sure that all conditions are valid."
                                ),
                            )
                            all_good = False

                        all_good = self._validate_categorical_and_boolean_values_check(
                            predicate,
                            all_good=all_good,
                            step_id=step.id,
                            link_condition=link.condition,
                            flow_id=flow.id,
                        )

                if isinstance(step, CollectInformationFlowStep):
                    predicates = [predicate.if_ for predicate in step.rejections]
                    for predicate in predicates:
                        all_good = self._verify_namespaces(
                            predicate, step.id, flow.id, all_good
                        )

                        pred, all_good = self._construct_predicate(
                            predicate, step.id, context, is_step=True, all_good=all_good
                        )
                        if pred and not pred.is_valid():
                            structlogger.error(
                                "validator.verify_predicates.invalid_rejection",
                                step=step.id,
                                rejection=predicate,
                                flow=flow.id,
                                event_info=(
                                    f"Detected invalid rejection '{predicate}' "
                                    f"at `collect` step '{step.id}' "
                                    f"for flow id '{flow.id}'. "
                                    f"Please make sure that all conditions are valid."
                                ),
                            )
                            all_good = False
        return all_good

    def _verify_namespaces(
        self, predicate: str, step_id: str, flow_id: str, all_good: bool
    ) -> bool:
        slots = re.findall(r"\bslots\.\w+", predicate)
        results: List[bool] = [all_good]

        if slots:
            domain_slots = {slot.name: slot for slot in self.domain.slots}
            for slot in slots:
                slot_name = slot.split(".")[1]
                if slot_name not in domain_slots:
                    structlogger.error(
                        "validator.verify_namespaces.invalid_slot",
                        slot=slot_name,
                        step=step_id,
                        flow=flow_id,
                        event_info=(
                            f"Detected invalid slot '{slot_name}' "
                            f"at step '{step_id}' "
                            f"for flow id '{flow_id}'. "
                            f"Please make sure that all slots are specified "
                            f"in the domain file."
                        ),
                    )
                    results.append(False)

        if not slots:
            # no slots found, check if context namespace is used
            variables = re.findall(r"\bcontext\.\w+", predicate)
            if not variables:
                structlogger.error(
                    "validator.verify_namespaces"
                    ".referencing_variables_without_namespace",
                    step=step_id,
                    predicate=predicate,
                    flow=flow_id,
                    event_info=(
                        f"Predicate '{predicate}' at step '{step_id}' for flow id "
                        f"'{flow_id}' references one or more variables  without "
                        f"the `slots.` or `context.` namespace prefix. "
                        f"Please make sure that all variables reference the required "
                        f"namespace."
                    ),
                )
                results.append(False)

        return all(results)

    def verify_flows(self) -> bool:
        """Checks for inconsistencies across flows."""
        structlogger.info("validation.flows.started")

        if self.flows.is_empty():
            structlogger.warn(
                "validator.verify_flows",
                event_info=(
                    "No flows were found in the data files. "
                    "Will not proceed with flow validation."
                ),
            )
            return True

        # add all flow validation conditions here
        flow_validation_conditions = [
            self.verify_flows_steps_against_domain(),
            self.verify_unique_flows(),
            self.verify_predicates(),
            self.verify_slot_persistence_configuration(),
        ]

        all_good = all(flow_validation_conditions)

        structlogger.info("validation.flows.ended")

        return all_good

    def _get_response_translation_warnings(self) -> list:
        """Collect warnings for responses missing translations.

        Returns:
            List of warnings for responses missing translations.
        """
        additional_languages = self.config.get(CONFIG_ADDITIONAL_LANGUAGES_KEY) or []
        response_warnings = []

        for response_name, responses in self.domain.responses.items():
            provided_languages = set()
            # For each response variation, we check if the additional
            # languages are available in at least on variation
            for response in responses:
                translation = response.get(KEY_TRANSLATION) or {}
                for language_code in additional_languages:
                    if translation.get(language_code):
                        provided_languages.add(language_code)

            missing_languages = [
                lang for lang in additional_languages if lang not in provided_languages
            ]
            if missing_languages:
                language_code_str = ", ".join(missing_languages)
                response_warnings.append(
                    {
                        "event": (
                            "validator.verify_translations.missing_response_translation"
                        ),
                        "response": response_name,
                        "missing_languages": missing_languages,
                        "event_info": (
                            f"The response '{response_name}' is "
                            f"missing a translation for the following "
                            f"languages: {language_code_str}."
                        ),
                    }
                )
        return response_warnings

    def _get_flow_translation_warnings(self) -> list:
        """Collect warnings for flows missing translations.

        Returns:
            List of warnings for flows missing translations.
        """
        additional_languages = self.config.get(CONFIG_ADDITIONAL_LANGUAGES_KEY) or []

        flow_warnings = []
        for flow in self.flows.underlying_flows:
            required_field_translation = [KEY_NAME]
            missing_languages = []
            for language_code in additional_languages:
                translation = flow.translation.get(language_code)
                # If translation for the language code doesn't exist,
                # or the required fields are not set properly,
                # we add the language code to the list.
                if not translation or not all(
                    getattr(translation, field, None)
                    for field in required_field_translation
                ):
                    missing_languages.append(language_code)

            if missing_languages:
                language_code_str = ", ".join(missing_languages)
                flow_warnings.append(
                    {
                        "event": (
                            "validator.verify_translations.missing_flow_translation"
                        ),
                        "flow": flow.id,
                        "missing_languages": missing_languages,
                        "event_info": (
                            f"The flow '{flow.id}' is missing the translation for "
                            f"the following languages: {language_code_str}."
                        ),
                    }
                )
        return flow_warnings

    def verify_config_language(self) -> bool:
        """Verify that config languages are properly set up.

        Returns:
            `True` if all languages are properly set up, `False` otherwise.

        Raises:
            RasaException: If the default language is listed as an
            additional language or if the language code is invalid.
        """
        language = self.config.get(CONFIG_LANGUAGE_KEY)
        additional_languages = self.config.get(CONFIG_ADDITIONAL_LANGUAGES_KEY, [])

        # Check if the default language is in the additional languages.
        if language in additional_languages:
            raise RasaException(
                f"The default language '{language}' is listed as an additional "
                f"language in the configuration file. Please remove it from "
                f"the list of additional languages."
            )

        # Verify the language codes by initializing the Language class.
        for language_code in [language] + additional_languages:
            Language.from_language_code(language_code=language_code)

        return True

    def verify_translations(self, summary_mode: bool = False) -> bool:
        """Checks for inconsistencies in translations.

        Args:
            summary_mode: If True, logs a single aggregated warning per category;
                otherwise, logs each warning individually.

        Returns:
            `True` if no inconsistencies were found, `False` otherwise.

        Raises:
            Warning: Single warning per response or flow missing translations
            if `summary_mode` is `True`, otherwise one warning per missing translation.
        """
        all_good = self.verify_config_language()

        additional_languages = self.config.get(CONFIG_ADDITIONAL_LANGUAGES_KEY, [])
        if not additional_languages:
            return all_good

        response_warnings = self._get_response_translation_warnings()
        flow_warnings = self._get_flow_translation_warnings()

        if summary_mode:
            if response_warnings:
                count = len(response_warnings)
                structlogger.warn(
                    "validator.verify_translations.missing_response_translation_summary",
                    count=count,
                    event_info=(
                        f"{count} response{' is' if count == 1 else 's are'} "
                        f"missing translations for some languages. "
                        "Run 'rasa data validate translations' for details."
                    ),
                )
            if flow_warnings:
                count = len(flow_warnings)
                structlogger.warn(
                    "validator.verify_translations.missing_flow_translation_summary",
                    count=count,
                    event_info=(
                        f"{count} flow{' is' if count == 1 else 's are'} "
                        f"missing translations for some languages. "
                        "Run 'rasa data validate translations' for details."
                    ),
                )
        else:
            for warning in response_warnings + flow_warnings:
                structlogger.warn(**warning)

        return all_good

    def validate_button_payloads(self) -> bool:
        """Check if the response button payloads are valid."""
        all_good = True
        for utter_name, response in self.domain.responses.items():
            for variation in response:
                for button in variation.get("buttons", []):
                    payload = button.get("payload")
                    if payload is None:
                        structlogger.error(
                            "validator.validate_button_payloads.missing_payload",
                            event_info=(
                                f"The button '{button.get('title')}' in response "
                                f"'{utter_name}' does not have a payload. "
                                f"Please add a payload to the button."
                            ),
                        )
                        all_good = False
                        continue

                    if not payload.strip():
                        structlogger.error(
                            "validator.validate_button_payloads.empty_payload",
                            event_info=(
                                f"The button '{button.get('title')}' in response "
                                f"'{utter_name}' has an empty payload. "
                                f"Please add a payload to the button."
                            ),
                        )
                        all_good = False
                        continue

                    regex_reader = create_regex_pattern_reader(
                        UserMessage(text=payload), self.domain
                    )

                    if regex_reader is None:
                        structlogger.warning(
                            "validator.validate_button_payloads.free_form_string",
                            event_info=(
                                "Using a free form string in payload of a button "
                                "implies that the string will be sent to the NLU "
                                "interpreter for parsing. To avoid the need for "
                                "parsing at runtime, it is recommended to use one "
                                "of the documented formats "
                                "(https://rasa.com/docs/rasa-pro/concepts/responses#buttons)"
                            ),
                        )
                        continue

                    if isinstance(
                        regex_reader, CommandPayloadReader
                    ) and regex_reader.is_above_slot_limit(payload):
                        structlogger.error(
                            "validator.validate_button_payloads.slot_limit_exceeded",
                            event_info=(
                                f"The button '{button.get('title')}' in response "
                                f"'{utter_name}' has a payload that sets more than "
                                f"{MAX_NUMBER_OF_SLOTS} slots. "
                                f"Please make sure that the number of slots set by "
                                f"the button payload does not exceed the limit."
                            ),
                        )
                        all_good = False
                        continue

                    if isinstance(regex_reader, YAMLStoryReader):
                        # the payload could contain double curly braces
                        # we need to remove 1 set of curly braces
                        payload = payload.replace("{{", "{").replace("}}", "}")

                    resulting_message = regex_reader.unpack_regex_message(
                        message=Message(data={"text": payload}), domain=self.domain
                    )

                    if not (
                        resulting_message.has_intent()
                        or resulting_message.has_commands()
                    ):
                        structlogger.error(
                            "validator.validate_button_payloads.invalid_payload_format",
                            event_info=(
                                f"The button '{button.get('title')}' in response "
                                f"'{utter_name}' does not follow valid payload formats "
                                f"for triggering a specific intent and entities or for "
                                f"triggering a SetSlot command."
                            ),
                            calm_docs_link=DOCS_URL_RESPONSES + "#payload-syntax",
                            nlu_docs_link=DOCS_URL_RESPONSES
                            + "#triggering-intents-or-passing-entities",
                        )
                        all_good = False

                    if resulting_message.has_commands():
                        # validate that slot names are unique
                        slot_names = set()
                        for command in resulting_message.get(COMMANDS, []):
                            slot_name = command.get("name")
                            if slot_name and slot_name in slot_names:
                                structlogger.error(
                                    "validator.validate_button_payloads.duplicate_slot_name",
                                    event_info=(
                                        f"The button '{button.get('title')}' "
                                        f"in response '{utter_name}' has a "
                                        f"command to set the slot "
                                        f"'{slot_name}' multiple times. "
                                        f"Please make sure that each slot "
                                        f"is set only once."
                                    ),
                                )
                                all_good = False
                            slot_names.add(slot_name)

        return all_good

    def validate_CALM_slot_mappings(self) -> bool:
        """Check if the usage of slot mappings in a CALM assistant is valid."""
        all_good = True

        for slot in self.domain._user_slots:
            nlu_mappings = any(
                [mapping.type.is_predefined_type() for mapping in slot.mappings]
            )
            llm_mappings = any(
                [mapping.type == SlotMappingType.FROM_LLM for mapping in slot.mappings]
            )
            controlled_mappings = any(
                [
                    mapping.type == SlotMappingType.CONTROLLED
                    for mapping in slot.mappings
                ]
            )

            all_good = self._allow_nlu_correction_is_valid(slot, nlu_mappings, all_good)

            all_good = self._custom_action_name_is_defined_in_the_domain(
                controlled_mappings, slot, all_good
            )

            all_good = self._validate_controlled_mappings(slot, all_good)

            all_good = self._config_contains_nlu_command_adapter(
                nlu_mappings, slot.name, all_good
            )

            all_good = self._uses_from_llm_mappings_in_a_NLU_based_assistant(
                llm_mappings, slot.name, all_good
            )

        return all_good

    @staticmethod
    def _allow_nlu_correction_is_valid(
        slot: Slot, nlu_mappings: bool, all_good: bool
    ) -> bool:
        """Verify that `allow_nlu_correction` property is used correctly in a `from_llm` mappings only."""  # noqa: E501
        if not slot.mappings:
            return all_good

        invalid_usage = False

        for mapping in slot.mappings:
            allow_nlu_correction = mapping.allow_nlu_correction
            if allow_nlu_correction and mapping.type != SlotMappingType.FROM_LLM:
                invalid_usage = True

            if allow_nlu_correction and not nlu_mappings:
                structlogger.error(
                    "validator.validate_slot_mappings_in_CALM.nlu_mappings_not_present",
                    slot_name=slot.name,
                    event_info=(
                        f"The slot '{slot.name}' does not have any "
                        f"NLU-based slot mappings. "
                        f"The property `allow_nlu_correction` is only "
                        f"applicable when the slot "
                        f"contains both NLU-based and LLM-based slot mappings."
                    ),
                )
                all_good = False

        if invalid_usage:
            structlogger.error(
                "validator.validate_slot_mappings_in_CALM.allow_nlu_correction",
                slot_name=slot.name,
                event_info=(
                    f"The slot '{slot.name}' has at least 1 slot mapping with "
                    f"'{KEY_ALLOW_NLU_CORRECTION}' set to 'true', but "
                    f"the slot mapping type is not 'from_llm'. "
                    f"Please set the slot mapping type to 'from_llm' "
                    f"to allow the LLM to correct this slot."
                ),
            )
            all_good = False

        return all_good

    def _custom_action_name_is_defined_in_the_domain(
        self,
        controlled_mappings: bool,
        slot: Slot,
        all_good: bool,
    ) -> bool:
        if not controlled_mappings:
            return all_good

        for mapping in slot.mappings:
            if (
                mapping.run_action_every_turn is not None
                and mapping.run_action_every_turn
                not in self.domain.action_names_or_texts
            ):
                structlogger.error(
                    "validator.validate_slot_mappings_in_CALM.custom_action_not_in_domain",
                    slot_name=slot.name,
                    action_name=mapping.run_action_every_turn,
                    event_info=(
                        f"The slot '{slot.name}' has a custom action "
                        f"'{mapping.run_action_every_turn}' "
                        f"defined in its slot mappings, "
                        f"but the action is not listed in the domain actions. "
                        f"Please add the action to your domain file."
                    ),
                )
                all_good = False

        return all_good

    def _config_contains_nlu_command_adapter(
        self, nlu_mappings: bool, slot_name: str, all_good: bool
    ) -> bool:
        if not nlu_mappings:
            return all_good

        if not self.flows:
            return all_good

        contains_nlu_command_adapter = any(
            [
                component.get("name") == "NLUCommandAdapter"
                for component in self.config.get(CONFIG_PIPELINE_KEY, [])
            ]
        )

        if not contains_nlu_command_adapter:
            structlogger.error(
                "validator.validate_slot_mappings_in_CALM.nlu_mappings_without_adapter",
                slot_name=slot_name,
                event_info=(
                    f"The slot '{slot_name}' has NLU slot mappings, "
                    f"but the NLUCommandAdapter is not present in the "
                    f"pipeline. Please add the NLUCommandAdapter to the "
                    f"pipeline in the config file."
                ),
                docs_link=DOCS_URL_DOMAIN + "#nlu-based-predefined-slot-mappings",
            )
            all_good = False

        return all_good

    def _uses_from_llm_mappings_in_a_NLU_based_assistant(
        self, llm_mappings: bool, slot_name: str, all_good: bool
    ) -> bool:
        if not llm_mappings:
            return all_good

        if self.flows:
            return all_good

        structlogger.error(
            "validator.validate_slot_mappings_in_CALM.llm_mappings_without_flows",
            slot_name=slot_name,
            event_info=(
                f"The slot '{slot_name}' has LLM slot mappings, "
                f"but no flows are present in the training data files. "
                f"Please add flows to the training data files."
            ),
        )
        return False

    def verify_slot_persistence_configuration(self) -> bool:
        """Verifies the validity of slot persistence after flow ends configuration.

        Only slots used in either a collect step or a set_slot step can be persisted and
        the configuration can either set at the flow level or the collect step level,
        but not both.

        Returns:
            bool: True if all slot persistence configuration is valid, False otherwise.

        Raises:
            DeprecationWarning: If reset_after_flow_ends is used in collect steps.
        """
        all_good = True

        for flow in self.flows.underlying_flows:
            flow_id = flow.id
            persist_slots = flow.persisted_slots
            has_flow_level_persistence = True if persist_slots else False
            flow_slots = set()

            for step in flow.steps_with_calls_resolved:
                if isinstance(step, SetSlotsFlowStep):
                    flow_slots.update([slot["key"] for slot in step.slots])

                elif isinstance(step, CollectInformationFlowStep):
                    collect_step = step.collect
                    flow_slots.add(collect_step)
                    if not step.reset_after_flow_ends:
                        warn_deprecated_collect_step_config()

                        if has_flow_level_persistence:
                            structlogger.error(
                                "validator.verify_slot_persistence_configuration.duplicate_config",
                                flow=flow_id,
                                collect_step=collect_step,
                                event_info=get_duplicate_slot_persistence_config_error_message(
                                    flow_id, collect_step
                                ),
                            )
                            all_good = False

            if has_flow_level_persistence:
                if not self._is_persist_slots_valid(persist_slots, flow_slots, flow_id):
                    all_good = False
        return all_good

    def _is_persist_slots_valid(
        self, persist_slots: List[str], flow_slots: Set[str], flow_id: str
    ) -> bool:
        invalid_slots = set(persist_slots) - flow_slots
        is_valid = False if invalid_slots else True

        if invalid_slots:
            structlogger.error(
                "validator.verify_slot_persistence_configuration.invalid_persist_slot",
                flow=flow_id,
                event_info=get_invalid_slot_persistence_config_error_message(
                    flow_id, invalid_slots
                ),
            )
        return is_valid

    def verify_studio_supported_validations(self) -> bool:
        """Validates the assistant project for Rasa Studio supported features.

        Ensure to add new validations here if they are required for
        Rasa Studio Upload CLI.
        """
        if self.domain.is_empty():
            structlogger.error(
                "rasa.validator.verify_studio_supported_validations.empty_domain",
                event_info="Encountered empty domain during validation.",
            )
            raise ValidationError(
                code="rasa.validator.verify_studio_supported_validations.empty_domain",
                event_info="Encountered empty domain during validation.",
            )

        self.warn_if_config_mandatory_keys_are_not_set()

        valid_responses = (
            self.check_for_no_empty_parenthesis_in_responses()
            and self.validate_button_payloads()
        )
        valid_nlu = self.verify_nlu()
        valid_flows = all(
            [
                self.verify_flows_steps_against_domain(),
                self.verify_unique_flows(),
                self.verify_predicates(),
            ]
        )
        valid_translations = self.verify_translations(summary_mode=True)
        valid_calm_slot_mappings = self.validate_CALM_slot_mappings()

        all_good = (
            valid_responses
            and valid_nlu
            and valid_flows
            and valid_translations
            and valid_calm_slot_mappings
        )

        return all_good

    def verify_slot_validation(self) -> bool:
        """Validates the slot validation configuration in the domain file."""
        all_good = True

        for slot in self.domain._user_slots:
            if slot.requires_validation():
                refill_utter = slot.validation.refill_utter  # type: ignore[union-attr]
                if refill_utter and refill_utter not in self.domain.responses:
                    self._log_slot_validation_error(
                        slot.name, "refill utterance", refill_utter
                    )
                    all_good = False
                rejections = slot.validation.rejections  # type: ignore[union-attr]
                for rejection in rejections:
                    if rejection.utter not in self.domain.responses:
                        self._log_slot_validation_error(
                            slot.name, "rejection utterance", rejection.utter
                        )
                        all_good = False

        return all_good

    def _log_slot_validation_error(self, slot_name: str, key: str, value: str) -> None:
        structlogger.error(
            "validator.verify_slot_validation.response_not_in_domain",
            slot=slot_name,
            event_info=(
                f"The slot '{slot_name}' requires validation, "
                f"but the {key} '{value}' "
                f"is not listed in the domain responses. "
                f"Please add it to your domain file."
            ),
        )

    @staticmethod
    def _validate_controlled_mappings(slot: Slot, all_good: bool) -> bool:
        for mapping in slot.mappings:
            if (
                mapping.run_action_every_turn is not None
                and mapping.type != SlotMappingType.CONTROLLED
            ):
                structlogger.error(
                    "validator.validate_slot_mappings_in_CALM.run_action_every_turn_invalid",
                    slot_name=slot.name,
                    event_info=(
                        f"The slot '{slot.name}' has a custom action "
                        f"'{mapping.run_action_every_turn}' "
                        f"defined in its slot mapping, "
                        f"but the slot mapping type is not 'controlled'. "
                    ),
                )
                all_good = False

            if (
                mapping.coexistence_system is not None
                and mapping.type != SlotMappingType.CONTROLLED
            ):
                structlogger.error(
                    "validator.validate_slot_mappings_in_CALM.coexistence_system_invalid",
                    slot_name=slot.name,
                    event_info=(
                        f"The slot '{slot.name}' has a coexistence system "
                        f"'{mapping.coexistence_system.value}' "
                        f"defined in its slot mapping, "
                        f"but the slot mapping type is not 'controlled'. "
                    ),
                )
                all_good = False

            if (
                mapping.coexistence_system is not None
                and mapping.coexistence_system != CoexistenceSystemType.SHARED
                and slot.shared_for_coexistence
            ):
                structlogger.error(
                    "validator.validate_slot_mappings_in_CALM.shared_for_coexistence_invalid",
                    slot_name=slot.name,
                    event_info=(
                        f"The slot '{slot.name}' has the `shared_for_coexistence` "
                        f"property set to `True`, but the slot mapping `controlled` "
                        f"type defines the `coexistence_system` property with a "
                        f"value different to the expected `SHARED` value. "
                    ),
                )
                all_good = False

        multiple_controlled_mappings = {
            mapping.coexistence_system.value
            for mapping in slot.mappings
            if mapping.type == SlotMappingType.CONTROLLED
            and mapping.coexistence_system is not None
        }
        contains_inconsistent_coexistence_system = len(multiple_controlled_mappings) > 1

        if contains_inconsistent_coexistence_system:
            structlogger.error(
                "validator.validate_slot_mappings_in_CALM.inconsistent_multiple_mappings",
                slot_name=slot.name,
                event_info=(
                    f"The slot '{slot.name}' has multiple `controlled` mappings "
                    f"with different coexistence systems defined: "
                    f"'{sorted(list(multiple_controlled_mappings))}'. "
                    f"Please only define one coexistence system for the slot. "
                ),
            )
            all_good = False

        return all_good

    def validate_conditional_response_variation_predicates(self) -> bool:
        """Validate the conditional response variation predicates."""
        context = {SLOTS: {slot.name: None for slot in self.domain.slots}}
        all_good = True

        for utter_name, variations in self.domain.responses.items():
            for variation in variations:
                condition = variation.get(RESPONSE_CONDITION)
                if not isinstance(condition, str):
                    continue

                predicate, all_good = self._construct_predicate(
                    condition,
                    utter_name,
                    context,
                    is_step=False,
                    all_good=all_good,
                )
                if not predicate:
                    continue

                if not predicate.is_valid():
                    structlogger.error(
                        "validator.validate_conditional_response_variation_predicates.invalid_condition",
                        utter=utter_name,
                        event_info=(
                            f"Detected invalid condition '{condition}' "
                            f"for response variation '{utter_name}'. "
                            f"Please make sure that all conditions are valid."
                        ),
                    )
                    all_good = False
                    continue

                predicate_syntax_tree = self._extract_predicate_syntax_tree(predicate)
                slot_namespace, _ = self._extract_slot_name_and_slot_value(
                    predicate_syntax_tree
                )

                if slot_namespace is not None and slot_namespace[0] != SLOTS:
                    structlogger.error(
                        "validator.validate_conditional_response_variation_predicates.invalid_namespace",
                        utter=utter_name,
                        event_info=(
                            f"Detected invalid namespace '{slot_namespace[0]}' in "
                            f"condition '{condition}' for response variation "
                            f"'{utter_name}'. Please make sure that you're "
                            f"using a valid namespace. "
                            f"The current supported option is: 'slots'."
                        ),
                    )
                    all_good = False
                    continue

                if (
                    slot_namespace is not None
                    and slot_namespace[1] not in context[SLOTS]
                ):
                    structlogger.error(
                        "validator.validate_conditional_response_variation_predicates.invalid_slot",
                        utter=utter_name,
                        event_info=(
                            f"Detected invalid slot '{slot_namespace[0]}' in "
                            f"condition '{condition}' for response variation "
                            f"'{utter_name}'. Please make sure that all slots "
                            f"are specified in the domain file."
                        ),
                    )
                    all_good = False

        return all_good



from __future__ import annotations

import dataclasses
from typing import Any, Dict, List, Text

import rasa.shared.utils.io
from rasa.engine.graph import ExecutionContext, GraphComponent
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.core.domain import Domain


@dataclasses.dataclass
class Responses:
    """Holds the responses of the domain."""

    data: Dict[Text, List[Dict[Text, Any]]]

    def fingerprint(self) -> Text:
        """Returns a fingerprint of the responses."""
        return rasa.shared.utils.io.get_dictionary_fingerprint(self.data)

    def get(self, key: Text, default: Any) -> Any:
        """Returns the value for the given key."""
        return self.data.get(key, default)


class ResponsesProvider(GraphComponent):
    """Provides responses during training and inference time."""

    @classmethod
    def create(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> ResponsesProvider:
        """Creates component (see parent class for full docstring)."""
        return cls()

    def provide(self, domain: Domain) -> Responses:
        """Returns the responses from the given domain."""
        return Responses(data=domain.responses)



from __future__ import annotations

import logging
import re
from typing import Any, Dict, List, Optional, Text, Tuple, Type

import numpy as np
import scipy.sparse

import rasa.nlu.utils.pattern_utils as pattern_utils
import rasa.shared.utils.io
import rasa.utils.io
from rasa.engine.graph import ExecutionContext, GraphComponent
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.nlu.constants import TOKENS_NAMES
from rasa.nlu.featurizers.sparse_featurizer.sparse_featurizer import SparseFeaturizer
from rasa.nlu.tokenizers.tokenizer import Tokenizer
from rasa.shared.nlu.constants import ACTION_TEXT, RESPONSE, TEXT
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.nlu.training_data.training_data import TrainingData

logger = logging.getLogger(__name__)


@DefaultV1Recipe.register(
    DefaultV1Recipe.ComponentType.MESSAGE_FEATURIZER, is_trainable=True
)
class RegexFeaturizer(SparseFeaturizer, GraphComponent):
    """Adds message features based on regex expressions."""

    @classmethod
    def required_components(cls) -> List[Type]:
        """Components that should be included in the pipeline before this component."""
        return [Tokenizer]

    @staticmethod
    def get_default_config() -> Dict[Text, Any]:
        """Returns the component's default config."""
        return {
            **SparseFeaturizer.get_default_config(),
            # text will be processed with case sensitive as default
            "case_sensitive": True,
            # use lookup tables to generate features
            "use_lookup_tables": True,
            # use regexes to generate features
            "use_regexes": True,
            # use match word boundaries for lookup table
            "use_word_boundaries": True,
        }

    def __init__(
        self,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        known_patterns: Optional[List[Dict[Text, Text]]] = None,
    ) -> None:
        """Constructs new features for regexes and lookup table using regex expressions.

        Args:
            config: Configuration for the component.
            model_storage: Storage which graph components can use to persist and load
                themselves.
            resource: Resource locator for this component which can be used to persist
                and load itself from the `model_storage`.
            execution_context: Information about the current graph run.
            known_patterns: Regex Patterns the component should pre-load itself with.
        """
        super().__init__(execution_context.node_name, config)

        self._model_storage = model_storage
        self._resource = resource

        self.known_patterns = known_patterns if known_patterns else []
        self.case_sensitive = config["case_sensitive"]
        self.finetune_mode = execution_context.is_finetuning

    @classmethod
    def create(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> RegexFeaturizer:
        """Creates a new untrained component (see parent class for full docstring)."""
        return cls(config, model_storage, resource, execution_context)

    def _merge_new_patterns(self, new_patterns: List[Dict[Text, Text]]) -> None:
        """Updates already known patterns with new patterns extracted from data.

        New patterns should always be added to the end of the existing
        patterns and the order of the existing patterns should not be disturbed.

        Args:
            new_patterns: Patterns extracted from training data and to be merged with
                known patterns.
        """
        pattern_name_index_map = {
            pattern["name"]: index for index, pattern in enumerate(self.known_patterns)
        }
        for extra_pattern in new_patterns:
            new_pattern_name = extra_pattern["name"]

            # Some patterns may have just new examples added
            # to them. These do not count as additional pattern.
            if new_pattern_name in pattern_name_index_map:
                self.known_patterns[pattern_name_index_map[new_pattern_name]][
                    "pattern"
                ] = extra_pattern["pattern"]
            else:
                self.known_patterns.append(extra_pattern)

    def train(self, training_data: TrainingData) -> Resource:
        """Trains the component with all patterns extracted from training data."""
        patterns_from_data = pattern_utils.extract_patterns(
            training_data,
            use_lookup_tables=self._config["use_lookup_tables"],
            use_regexes=self._config["use_regexes"],
            use_word_boundaries=self._config["use_word_boundaries"],
        )
        if self.finetune_mode:
            # Merge patterns extracted from data with known patterns
            self._merge_new_patterns(patterns_from_data)
        else:
            self.known_patterns = patterns_from_data

        self._persist()
        return self._resource

    def process_training_data(self, training_data: TrainingData) -> TrainingData:
        """Processes the training examples (see parent class for full docstring)."""
        for example in training_data.training_examples:
            for attribute in [TEXT, RESPONSE, ACTION_TEXT]:
                self._text_features_with_regex(example, attribute)

        return training_data

    def process(self, messages: List[Message]) -> List[Message]:
        """Featurizes all given messages in-place.

        Returns:
          the given list of messages which have been modified in-place
        """
        for message in messages:
            self._text_features_with_regex(message, TEXT)

        return messages

    def _text_features_with_regex(self, message: Message, attribute: Text) -> None:
        """Helper method to extract features and set them appropriately in the message.

        Args:
            message: Message to be featurized.
            attribute: Attribute of message to be featurized.
        """
        if self.known_patterns:
            sequence_features, sentence_features = self._features_for_patterns(
                message, attribute
            )

            self.add_features_to_message(
                sequence_features, sentence_features, attribute, message
            )

    def _features_for_patterns(
        self, message: Message, attribute: Text
    ) -> Tuple[Optional[scipy.sparse.coo_matrix], Optional[scipy.sparse.coo_matrix]]:
        """Checks which known patterns match the message.

        Given a sentence, returns a vector of {1,0} values indicating which
        regexes did match. Furthermore, if the
        message is tokenized, the function will mark all tokens with a dict
        relating the name of the regex to whether it was matched.

        Args:
            message: Message to be featurized.
            attribute: Attribute of message to be featurized.

        Returns:
           Token and sentence level features of message attribute.
        """
        # Attribute not set (e.g. response not present)
        if not message.get(attribute):
            return None, None

        tokens = message.get(TOKENS_NAMES[attribute], [])

        if not tokens:
            # nothing to featurize
            return None, None

        flags = 0  # default flag
        if not self.case_sensitive:
            flags = re.IGNORECASE

        sequence_length = len(tokens)

        num_patterns = len(self.known_patterns)

        sequence_features = np.zeros([sequence_length, num_patterns])
        sentence_features = np.zeros([1, num_patterns])

        for pattern_index, pattern in enumerate(self.known_patterns):
            matches = list(
                re.finditer(pattern["pattern"], message.get(attribute), flags=flags)
            )

            for token_index, t in enumerate(tokens):
                patterns = t.get("pattern", default={})
                patterns[pattern["name"]] = False

                for match in matches:
                    if t.start < match.end() and t.end > match.start():
                        patterns[pattern["name"]] = True
                        sequence_features[token_index][pattern_index] = 1.0
                        if attribute in [RESPONSE, TEXT, ACTION_TEXT]:
                            # sentence vector should contain all patterns
                            sentence_features[0][pattern_index] = 1.0

                t.set("pattern", patterns)

        return (
            scipy.sparse.coo_matrix(sequence_features),
            scipy.sparse.coo_matrix(sentence_features),
        )

    @classmethod
    def load(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        **kwargs: Any,
    ) -> RegexFeaturizer:
        """Loads trained component (see parent class for full docstring)."""
        known_patterns = None

        try:
            with model_storage.read_from(resource) as model_dir:
                patterns_file_name = model_dir / "patterns.json"
                known_patterns = rasa.shared.utils.io.read_json_file(patterns_file_name)
        except (ValueError, FileNotFoundError):
            logger.warning(
                f"Failed to load `{cls.__class__.__name__}` from model storage. "
                f"Resource '{resource.name}' doesn't exist."
            )

        return cls(
            config,
            model_storage,
            resource,
            execution_context,
            known_patterns=known_patterns,
        )

    def _persist(self) -> None:
        with self._model_storage.write_to(self._resource) as model_dir:
            regex_file = model_dir / "patterns.json"
            rasa.shared.utils.io.dump_obj_as_json_to_file(
                regex_file, self.known_patterns
            )

    @classmethod
    def validate_config(cls, config: Dict[Text, Any]) -> None:
        """Validates that the component is configured properly."""
        pass



from __future__ import annotations

import importlib
import inspect
import os
from dataclasses import fields, is_dataclass
from typing import Any, Dict, List, Optional, Set, Text

import importlib_resources
from pydantic import BaseModel, Field, ValidationInfo, field_validator

from rasa.dialogue_understanding.stack.frames.pattern_frame import PatternFlowStackFrame
from rasa.shared.core.constants import (
    DEFAULT_ACTION_NAMES,
    DEFAULT_INTENTS,
    RULE_SNIPPET_ACTION_NAME,
)
from rasa.shared.core.slots import AnySlot, BooleanSlot, CategoricalSlot, TextSlot

ACTIONS: List[Text] = [
    *DEFAULT_ACTION_NAMES,
    "validate_{{context.collect}}",
]
EXCLUDED_ACTIONS: Set[Text] = {RULE_SNIPPET_ACTION_NAME}

INTENTS: List[Text] = [*DEFAULT_INTENTS]

CONTEXT_FIELD_TYPES: Dict[Text, Text] = {
    "canceled_name": CategoricalSlot.type_name,
    "names": CategoricalSlot.type_name,
    "collect": CategoricalSlot.type_name,
    "utter": CategoricalSlot.type_name,
    "collect_action": CategoricalSlot.type_name,
    "previous_flow_name": CategoricalSlot.type_name,
    "reset_flow_id": CategoricalSlot.type_name,
    "is_reset_only": BooleanSlot.type_name,
    "canceled_frames": AnySlot.type_name,
    "corrected_slots": AnySlot.type_name,
    "rejections": AnySlot.type_name,
    "info": AnySlot.type_name,
    "reason": TextSlot.type_name,
    "error_type": TextSlot.type_name,
}

FLOW_NAME_VALUES = "FLOW_NAME"
FLOW_ID_VALUES = "FLOW_ID"
SLOT_NAME_VALUES = "SLOT_NAME"
RESPONSE_NAME_VALUES = "RESPONSE_NAME"
ACTION_NAME_VALUES = "ACTION_NAME"

CONTEXT_FIELD_VALUES: Dict[Text, Text] = {
    "canceled_name": FLOW_NAME_VALUES,
    "names": FLOW_NAME_VALUES,
    "previous_flow_name": FLOW_NAME_VALUES,
    "reset_flow_id": FLOW_ID_VALUES,
    "collect": SLOT_NAME_VALUES,
    "utter": RESPONSE_NAME_VALUES,
    "collect_action": ACTION_NAME_VALUES,
}

PATTERNS_MODULE_BASE = "rasa.dialogue_understanding.patterns"


class ContextField(BaseModel):
    """Element in the `contexts` mapping of the domain."""

    patterns: List[Text] = Field(
        ..., description="Patterns that reference this context field."
    )
    type: Text = Field(
        ..., description="Slot type (categorical, text, boolean, any …)."
    )
    values: Optional[List[Text]] = Field(
        None,
        description="Optional placeholder that restricts which values a slot can take "
        "(FLOW_NAME, SLOT_NAME, …).",
    )

    @field_validator("type")
    def _validate_slot_type(cls, v: str) -> str:
        allowed_types = list(set(CONTEXT_FIELD_TYPES.values()))
        if v not in allowed_types:
            raise ValueError(
                f"Unsupported type '{v}'. "
                f"Must be one of: {', '.join(allowed_types)}."
            )
        return v

    @field_validator("values")
    def _validate_values_placeholder(
        cls, v: Optional[str], values: ValidationInfo
    ) -> Optional[str]:
        if v is None:
            return v

        allowed_values = set(CONTEXT_FIELD_VALUES.values())
        if not set(v).issubset(allowed_values):
            raise ValueError(
                f"Unsupported values placeholder '{v}'. "
                f"Must be one of {', '.join(allowed_values)}."
            )

        slot_type = values.data.get("type")
        if slot_type != CategoricalSlot.type_name:
            raise ValueError(
                "`values` can only be specified for categorical slots "
                f"(got slot type '{slot_type}')."
            )

        return v


class PatternDomain(BaseModel):
    """Complete domain that is generated for the default patterns."""

    actions: List[Text]
    intents: List[Text]
    contexts: Dict[Text, ContextField]


def build_contexts_from_patterns() -> Dict[str, Dict[str, Any]]:
    """Builds a dictionary of contexts from the pattern classes.

    Returns:
        A dictionary where each key is a field name and the value is a dictionary
    """
    patterns_folder = str(importlib_resources.files(PATTERNS_MODULE_BASE))
    contexts_map: Dict[str, Dict[str, Any]] = {}

    # Dynamically gather all .py files
    for root, _, files in os.walk(patterns_folder):
        for file in files:
            if not file.endswith(".py"):
                continue

            try:
                module_name = f"{PATTERNS_MODULE_BASE}.{file[:-3]}"
                module = importlib.import_module(module_name)
            except ImportError:
                continue

            # Inspect classes in that module
            for _, cls in inspect.getmembers(module, inspect.isclass):
                if not is_dataclass(cls):
                    continue

                # The cls has to be a subclass of PatternFlowStackFrame
                if cls == PatternFlowStackFrame or not issubclass(
                    cls, PatternFlowStackFrame
                ):
                    continue

                for f in fields(cls):
                    field_name = f.name
                    if field_name not in contexts_map:
                        field_type = CONTEXT_FIELD_TYPES.get(
                            field_name, TextSlot.type_name
                        )
                        contexts_map[field_name] = {
                            "patterns": set(),
                            "type": field_type,
                        }
                        values: Optional[Text] = CONTEXT_FIELD_VALUES.get(field_name)
                        if values:
                            contexts_map[field_name]["values"] = [values]

                    # Add the pattern name to the set
                    pattern_name = cls.type()
                    contexts_map[field_name]["patterns"].add(pattern_name)

    # Convert "patterns" from set to list, for a clean final structure
    for field_name, details in contexts_map.items():
        details["patterns"] = sorted(list(details["patterns"]))

    return contexts_map


def generate_domain_for_default_patterns() -> PatternDomain:
    """
    Generate the domain for pattern-based flows as a strongly-typed object.

    Returns: PatternDomain Pydantic model containing actions, intents and contexts.
    """
    actions = [action for action in ACTIONS if action not in EXCLUDED_ACTIONS]
    raw_contexts = build_contexts_from_patterns()
    contexts = {
        field_name: ContextField(**details)
        for field_name, details in raw_contexts.items()
    }
    return PatternDomain(actions=actions, intents=INTENTS, contexts=contexts)



import json
from decimal import Decimal
from typing import Any, Text


class DecimalEncoder(json.JSONEncoder):
    """`json.JSONEncoder` that dumps `Decimal`s as `float`s."""

    def default(self, obj: Any) -> Any:
        """Get serializable object for `o`.

        Args:
            obj: Object to serialize.

        Returns:
            `obj` converted to `float` if `o` is a `Decimals`, else the base class
            `default()` method.
        """
        if isinstance(obj, Decimal):
            return float(obj)
        return super().default(obj)


class SetEncoder(json.JSONEncoder):
    """`json.JSONEncoder` that dumps `set`s as `list`s."""

    def default(self, obj: Any) -> Any:
        if isinstance(obj, set):
            return list(obj)
        return super().default(obj)


def replace_floats_with_decimals(obj: Any, round_digits: int = 9) -> Any:
    """Convert all instances in `obj` of `float` to `Decimal`.

    Args:
        obj: Input object.
        round_digits: Rounding precision of `Decimal` values.

    Returns:
        Input `obj` with all `float` types replaced by `Decimal`s rounded to
        `round_digits` decimal places.
    """

    def _float_to_rounded_decimal(s: Text) -> Decimal:
        return Decimal(s).quantize(Decimal(10) ** -round_digits)

    return json.loads(json.dumps(obj), parse_float=_float_to_rounded_decimal)


def replace_decimals_with_floats(obj: Any) -> Any:
    """Convert all instances in `obj` of `Decimal` to `float`.

    Args:
        obj: A `List` or `Dict` object.

    Returns:
        Input `obj` with all `Decimal` types replaced by `float`s.
    """
    return json.loads(json.dumps(obj, cls=DecimalEncoder))



DEFAULT_SERVER_PORT = 5005

DEFAULT_SERVER_INTERFACE = "0.0.0.0"

DEFAULT_SERVER_FORMAT = "{}://localhost:{}"

DEFAULT_SERVER_URL = DEFAULT_SERVER_FORMAT.format("http", DEFAULT_SERVER_PORT)

DEFAULT_INTERACTIVE_SERVER_URL = "{}://localhost:{}"

DEFAULT_NLU_FALLBACK_THRESHOLD = 0.3

DEFAULT_NLU_FALLBACK_AMBIGUITY_THRESHOLD = 0.1

DEFAULT_CORE_FALLBACK_THRESHOLD = 0.3

DEFAULT_MAX_HISTORY = None  # Core policy history is unbounded by default.

DEFAULT_RESPONSE_TIMEOUT = 60 * 60  # 1 hour

DEFAULT_REQUEST_TIMEOUT = 60 * 5  # 5 minutes

DEFAULT_STREAM_READING_TIMEOUT = 10  # in seconds

DEFAULT_LOCK_LIFETIME = 60  # in seconds

DEFAULT_KEEP_ALIVE_TIMEOUT = 120  # in seconds

BEARER_TOKEN_PREFIX = "Bearer "

# The lowest priority is intended to be used by machine learning policies.
DEFAULT_POLICY_PRIORITY = 1

# The priority of intent-prediction policies.
# This should be below all rule based policies but higher than ML
# based policies. This enables a loop inside ensemble where if none
# of the rule based policies predict an action and intent prediction
# policy predicts one, its prediction is chosen by the ensemble and
# then the ML based policies are again run to get the prediction for
# an actual action. To prevent an infinite loop, intent prediction
# policies only predict an action if the last event in
# the tracker is of type `UserUttered`. Hence, they make at most
# one action prediction in each conversation turn. This allows other
# policies to predict a winning action prediction.
UNLIKELY_INTENT_POLICY_PRIORITY = DEFAULT_POLICY_PRIORITY + 1

# The priority intended to be used by memoization policies.
# It is higher than default to prioritize training stories.
MEMOIZATION_POLICY_PRIORITY = UNLIKELY_INTENT_POLICY_PRIORITY + 1

# The priority of the `RulePolicy` is higher than all other policies since
# rule execution takes precedence over training stories or predicted actions.
RULE_POLICY_PRIORITY = MEMOIZATION_POLICY_PRIORITY + 1

# chat policy priority
CHAT_POLICY_PRIORITY = RULE_POLICY_PRIORITY + 1

# search policy priority
SEARCH_POLICY_PRIORITY = CHAT_POLICY_PRIORITY + 1

# flow policy priority
FLOW_POLICY_PRIORITY = SEARCH_POLICY_PRIORITY + 1

DIALOGUE = "dialogue"

# RabbitMQ message property header added to events published using `rasa export`
RASA_EXPORT_PROCESS_ID_HEADER_NAME = "rasa-export-process-id"

# Name of the environment variable defining the PostgreSQL schema to access. See
# https://www.postgresql.org/docs/9.1/ddl-schemas.html for more details.
POSTGRESQL_SCHEMA = "POSTGRESQL_SCHEMA"

# Names of the environment variables defining PostgreSQL pool size and max overflow
POSTGRESQL_POOL_SIZE = "SQL_POOL_SIZE"
POSTGRESQL_MAX_OVERFLOW = "SQL_MAX_OVERFLOW"

# File names for testing
CONFUSION_MATRIX_STORIES_FILE = "story_confusion_matrix.png"
REPORT_STORIES_FILE = "story_report.json"
FAILED_STORIES_FILE = "failed_test_stories.yml"
SUCCESSFUL_STORIES_FILE = "successful_test_stories.yml"
STORIES_WITH_WARNINGS_FILE = "stories_with_warnings.yml"

POLICY_PRIORITY = "priority"
POLICY_FEATURIZER = "featurizer"
POLICY_MAX_HISTORY = "max_history"

DEFAULT_PROTOCOL = "UDP"
DEFAULT_SYSLOG_HOST = "localhost"
DEFAULT_SYSLOG_PORT = 514

COMPRESS_ACTION_SERVER_REQUEST_ENV_NAME = "COMPRESS_ACTION_SERVER_REQUEST"
DEFAULT_COMPRESS_ACTION_SERVER_REQUEST = False

# uses python like string formatting for the interpolation of templates
RASA_FORMAT_TEMPLATE_ENGINE = "format"

# uses jinja for the interpolation of templates
JINJA2_TEMPLATE_ENGINE = "jinja"

# default engine used if no engine is specified in the response
DEFAULT_TEMPLATE_ENGINE = RASA_FORMAT_TEMPLATE_ENGINE

# configuration parameter used to specify the template engine to use
# for a response
TEMPLATE_ENGINE_CONFIG_KEY = "template"

# metadata keys for bot utterance events
UTTER_SOURCE_METADATA_KEY = "utter_source"
DOMAIN_GROUND_TRUTH_METADATA_KEY = "domain_ground_truth"
ACTIVE_FLOW_METADATA_KEY = "active_flow"
STEP_ID_METADATA_KEY = "step_id"
KEY_IS_CALM_SYSTEM = "is_calm_system"
KEY_IS_COEXISTENCE_ASSISTANT = "is_coexistence_assistant"



from pathlib import Path
from typing import Any, ClassVar, Dict, List, Optional, Text, Union

import jsonschema
import ruamel.yaml.nodes as yaml_nodes
from ruamel import yaml as yaml

import rasa.shared
import rasa.shared.data
import rasa.shared.utils.io
from rasa.shared.core.flows.flow import DEFAULT_RUN_PATTERN_COMPLETED, Flow
from rasa.shared.core.flows.flow_step import FlowStep
from rasa.shared.core.flows.flows_list import FlowsList
from rasa.shared.core.flows.steps import CollectInformationFlowStep
from rasa.shared.core.flows.steps.collect import (
    DEFAULT_ASK_BEFORE_FILLING,
    DEFAULT_FORCE_SLOT_FILLING,
    DEFAULT_RESET_AFTER_FLOW_ENDS,
)
from rasa.shared.exceptions import RasaException, YamlException
from rasa.shared.utils.yaml import (
    dump_obj_as_yaml_to_string,
    is_key_in_yaml,
    read_yaml,
    validate_yaml_with_jsonschema,
)

FLOWS_SCHEMA_FILE = "shared/core/flows/flows_yaml_schema.json"
KEY_FLOWS = "flows"


class YAMLFlowsReader:
    """Class that reads flows information in YAML format."""

    expand_env_vars: ClassVar[bool] = True

    @classmethod
    def read_from_file(
        cls, filename: Union[Text, Path], add_line_numbers: bool = True
    ) -> FlowsList:
        """Read flows from file.

        Args:
            filename: Path to the flows file.
            add_line_numbers: Flag whether to add line numbers to yaml

        Returns:
            `Flow`s read from `filename`.
        """
        try:
            return cls.read_from_string(
                rasa.shared.utils.io.read_file(
                    filename, rasa.shared.utils.io.DEFAULT_ENCODING
                ),
                add_line_numbers=add_line_numbers,
                file_path=filename,
            )
        except YamlException as e:
            e.filename = str(filename)
            raise e
        except RasaException as e:
            raise YamlException(filename) from e

    @staticmethod
    def humanize_flow_error(error: jsonschema.ValidationError) -> str:
        """Create a human understandable error message from a validation error.

        Converts a jsonschema validation error into a human understandable
        error message. This is used to provide more helpful error messages
        when a user provides an invalid flow definition.

        The documentation for the `jsonschema.ValidationError` can be found
        here https://python-jsonschema.readthedocs.io/en/latest/errors/#best-match-and-relevance

        Args:
            error: The validation error to convert.

        Returns:
            A human understandable error message.
        """

        def faulty_property(path: List[Any]) -> str:
            """Get the name of the property that caused the error.

            The exception contains a path to the property that caused the error.
            We will use that path to get the name of the property.

            Example:
                > faulty_property(['flows', 'add_contact', 'steps', 0, 'next'])
                'next'

            Args:
                path: The path to the property that caused the error.

            Returns:
                The name of the property that caused the error.
            """
            if not path:
                return "schema"
            if isinstance(path[-1], int):
                # the path is pointing towards an element in a list, so
                # we use the name of the list if possible
                return path[-2] if len(path) > 1 else "list"
            return str(path[-1])

        def schema_name(schema: Dict[str, Any]) -> str:
            """Get the name of the schema.

            This helps when displaying error messages, as we don't want to
            show the schema itself, but rather a name that describes
            what we expect. E.g. the following schema

            ```
            "set_slots": {
                "type": "array",
                "schema_name": "list of slot sets",
                "items": {
                    "type": "object"
                }
            }
            ```
            has a `schema_name` set. When we need to raise an error because
            this schema was not satisified, we will use the `schema_name`
            instead of the type itself. The type is less specific (`array`)
            and therefore less usefull than the handcrafted `schema_name`.

            If a schema does not have a `schema_name` set, we will use the
            `type` instead as a fallback.
            """
            return schema.get("schema_name", schema.get("type"))

        def schema_names(schemas: List[Dict[str, Any]]) -> List[str]:
            """Get the names of the schemas.

            Example:
                > schema_names([
                    {"required": ["action"], "schema_name": "action step"},
                    {"required": ["collect"], "schema_name": "collect step"},
                    {"required": ["link"], "schema_name": "link step"},
                    {"required": ["set_slots"], "schema_name": "slot set step"},
                    {"required": ["noop"], "schema_name": ""}])
                ['action step', 'collect step', 'link step', 'slot set step']

            Args:
                schemas: The schemas to get the names of.

            Returns:
                The names of the schemas.
            """
            names = []
            for schema in schemas:
                if name := schema_name(schema):
                    names.append(name)
            return names

        def expected_schema(error: jsonschema.ValidationError, schema_type: str) -> str:
            """Get the expected schema."""
            expected_schemas = error.schema.get(schema_type, [])
            expected = schema_names(expected_schemas)
            if expected:
                return " or ".join(sorted(expected))
            else:
                return str(error.schema)

        def format_oneof_error(error: jsonschema.ValidationError) -> str:
            """Format a oneOf error."""
            return (
                f"Not a valid '{faulty_property(error.absolute_path)}' definition. "
                f"Expected {expected_schema(error, 'oneOf')}."
            )

        def format_anyof_error(error: jsonschema.ValidationError) -> str:
            """Format an anyOf error."""
            return (
                f"Not a valid '{faulty_property(error.absolute_path)}' definition. "
                f"Expected {expected_schema(error, 'anyOf')}."
            )

        def format_type_error(error: jsonschema.ValidationError) -> str:
            """Format a type error."""
            expected_value = schema_name(error.schema)
            if isinstance(error.instance, dict):
                instance = "a dictionary"
            elif isinstance(error.instance, list):
                instance = "a list"
            else:
                instance = f"`{error.instance}`"
            return f"Found {instance} but expected a {expected_value}."

        if error.validator == "oneOf":
            return format_oneof_error(error)

        if error.validator == "anyOf":
            return format_anyof_error(error)

        if error.validator == "type":
            return format_type_error(error)

        if error.validator == "additionalProperties":
            return error.message

        if error.validator == "required":
            return error.message

        return (
            f"The flow at {error.json_path} is not valid. "
            f"Please double check your flow definition."
        )

    @classmethod
    def read_from_string(
        cls,
        string: Text,
        add_line_numbers: bool = True,
        file_path: Optional[Union[str, Path]] = None,
    ) -> FlowsList:
        """Read flows from a string.

        Args:
            string: Unprocessed YAML file content.
            add_line_numbers: If true, a custom constructor is added to add line
                numbers to each node.
            file_path: File path of the flow.

        Returns:
            `Flow`s read from `string`.
        """
        validate_yaml_with_jsonschema(
            string,
            FLOWS_SCHEMA_FILE,
            humanize_error=cls.humanize_flow_error,
            expand_env_vars=cls.expand_env_vars,
        )
        if add_line_numbers:
            yaml_content = read_yaml(
                string,
                custom_constructor=line_number_constructor,
                expand_env_vars=cls.expand_env_vars,
            )
            yaml_content = process_yaml_content(yaml_content)

        else:
            yaml_content = read_yaml(string, expand_env_vars=cls.expand_env_vars)

        return FlowsList.from_json(yaml_content.get(KEY_FLOWS, {}), file_path=file_path)


class YamlFlowsWriter:
    """Class that writes flows information in YAML format."""

    @staticmethod
    def dumps(
        flows: List[Flow],
        should_clean_json: bool = False,
    ) -> Text:
        """Dump `Flow`s to YAML.

        Args:
            flows: The `Flow`s to dump.
            should_clean_json: Flag whether to clean the flow JSON.

        Returns:
            The dumped YAML.
        """
        dump = {}
        for flow in flows:
            dumped_flow = get_flow_as_json(flow, should_clean_json)
            del dumped_flow["id"]
            dump[flow.id] = dumped_flow
        return dump_obj_as_yaml_to_string({KEY_FLOWS: dump})

    @staticmethod
    def dump(
        flows: List[Flow],
        filename: Union[Text, Path],
        should_clean_json: bool = False,
    ) -> None:
        """Dump `Flow`s to YAML file.

        Args:
            flows: The `Flow`s to dump.
            filename: The path to the file to write to.
            should_clean_json: Flag whether to clean the flow JSON.
        """
        rasa.shared.utils.io.write_text_file(
            YamlFlowsWriter.dumps(flows, should_clean_json),
            filename,
        )


def is_flows_file(file_path: Union[Text, Path]) -> bool:
    """Check if file contains Flow training data.

    Args:
        file_path: Path of the file to check.

    Returns:
        `True` in case the file is a flows YAML training data file,
        `False` otherwise.

    Raises:
        YamlException: if the file seems to be a YAML file (extension) but
            can not be read / parsed.
    """
    return rasa.shared.data.is_likely_yaml_file(file_path) and is_key_in_yaml(
        file_path, KEY_FLOWS
    )


def line_number_constructor(loader: yaml.Loader, node: yaml_nodes.Node) -> Any:
    """A custom YAML constructor adding line numbers to nodes.

    Args:
        loader (yaml.Loader): The YAML loader.
        node (yaml.nodes.Node): The YAML node.

    Returns:
        Any: The constructed Python object with added line numbers in metadata.
    """
    if isinstance(node, yaml_nodes.MappingNode):
        mapping = loader.construct_mapping(node, deep=True)
        if "metadata" not in mapping:
            # We add the line information to the metadata of a flow step
            # Lines are 0-based index; adding 1 to start from
            # line 1 for human readability
            mapping["metadata"] = {
                "line_numbers": f"{node.start_mark.line + 1}-{node.end_mark.line}"
            }
        return mapping
    elif isinstance(node, yaml_nodes.SequenceNode):
        sequence = loader.construct_sequence(node, deep=True)
        for item in node.value:
            if isinstance(item, yaml_nodes.MappingNode):
                start_line = item.start_mark.line + 1
                end_line = item.end_mark.line
                # Only add line numbers to dictionary items within the sequence
                index = node.value.index(item)
                if isinstance(sequence[index], dict):
                    if "metadata" not in sequence[index]:
                        sequence[index]["metadata"] = {}
                    if "line_numbers" not in sequence[index]["metadata"]:
                        sequence[index]["metadata"]["line_numbers"] = (
                            f"{start_line}-{end_line}"
                        )

        return sequence
    return loader.construct_object(node, deep=True)


def _remove_keys_recursively(
    data: Union[Dict, List], keys_to_delete: List[str]
) -> None:
    """Recursively removes all specified keys in the given data.

    Special handling for 'metadata'.

    Args:
        data: The data structure (dictionary or list) to clean.
        keys_to_delete: A list of keys to remove from the dictionaries.
    """
    if isinstance(data, dict):
        keys = list(data.keys())
        for key in keys:
            if key in keys_to_delete:
                # Special case for 'metadata': only delete if it
                # only contains 'line_numbers'
                if key == "metadata" and isinstance(data[key], dict):
                    if len(data[key]) == 1 and "line_numbers" in data[key]:
                        del data[key]
                else:
                    del data[key]
            else:
                _remove_keys_recursively(data[key], keys_to_delete)
    elif isinstance(data, list):
        for item in data:
            _remove_keys_recursively(item, keys_to_delete)


def _process_keys_recursively(
    data: Union[Dict, List], keys_to_check: List[str]
) -> None:
    """Recursively iterates over YAML content and applies remove_keys_recursively."""
    if isinstance(data, dict):
        keys = list(
            data.keys()
        )  # Make a list of keys to avoid changing the dictionary size during iteration
        for key in keys:
            if key in keys_to_check:
                _remove_keys_recursively(data[key], ["metadata"])
            else:
                _process_keys_recursively(data[key], keys_to_check)
    elif isinstance(data, list):
        for item in data:
            _process_keys_recursively(item, keys_to_check)


def process_yaml_content(yaml_content: Dict[str, Any]) -> Dict[str, Any]:
    """Processes parsed YAML content to remove "metadata"."""
    # Remove metadata on the top level
    if "metadata" in yaml_content and (
        len(yaml_content["metadata"]) == 1
        and "line_numbers" in yaml_content["metadata"]
    ):
        del yaml_content["metadata"]

    # We expect metadata only under "flows" key...
    keys_to_delete_metadata = [key for key in yaml_content if key != "flows"]
    for key in keys_to_delete_metadata:
        _remove_keys_recursively(yaml_content[key], ["metadata"])

    # ...but "metadata" is also not a key of "flows"
    if "flows" in yaml_content and "metadata" in yaml_content["flows"]:
        if (
            len(yaml_content["flows"]["metadata"]) == 1
            and "line_numbers" in yaml_content["flows"]["metadata"]
        ):
            del yaml_content["flows"]["metadata"]

    # Under the "flows" key certain keys cannot have metadata
    _process_keys_recursively(
        yaml_content["flows"], ["nlu_trigger", "set_slots", "metadata"]
    )

    return yaml_content


def get_flow_as_json(flow: Flow, should_clean_json: bool = False) -> Dict[str, Any]:
    """
    Clean the Flow JSON by removing default values and empty fields.

    Args:
        flow: The Flow object to clean.
        should_clean_json: Flag indicating whether to clean the JSON.

    Returns:
        The cleaned Flow JSON as a dictionary.
    """
    step_id_to_default_id = {}
    flow_data = flow.as_json()
    if not should_clean_json:
        return flow_data

    def gather_step_ids(step: FlowStep) -> None:
        """Create a map of step IDs to their default IDs."""
        step_id_to_default_id[step.id] = step.default_id
        for link in step.next.links:
            for child in link.child_steps():
                gather_step_ids(child)

    def clean_flow_data(data: Dict[str, Any]) -> None:
        """Remove or omit flow-level fields if they match defaults."""
        for top_level_step in flow.step_sequence.steps:
            gather_step_ids(top_level_step)

        if data.get("run_pattern_completed") == DEFAULT_RUN_PATTERN_COMPLETED:
            data.pop("run_pattern_completed", None)

        if not data.get("persisted_slots"):
            data.pop("persisted_slots", None)

        _remove_keys_recursively(data, ["metadata"])

        data.pop("file_path", None)

        steps = data.get("steps", [])
        for i, current_step in enumerate(steps):
            # Look ahead to the next step if we should remove the 'next' field
            next_step_data = steps[i + 1] if i + 1 < len(steps) else None
            clean_step(current_step, next_step_data)

    def clean_step(current_step: Dict[str, Any], next_step: Dict[str, Any]) -> None:
        """Remove default fields from a step."""
        # Remove 'next' if it exactly matches the next step's default ID
        if next_step:
            next_id = next_step.get("id")
            default_id = step_id_to_default_id.get(next_id)
            if next_id and current_step.get("next") == default_id:
                current_step.pop("next", None)

        # Remove 'id' if it equals its own default
        step_id = current_step.get("id")
        if step_id and step_id == step_id_to_default_id.get(step_id):
            current_step.pop("id", None)

        if "collect" in current_step:
            clean_collect_step(current_step)

    def clean_collect_step(step_data: Dict[str, Any]) -> None:
        """Remove default fields from a collect step."""
        slot_name = step_data["collect"]
        default_utter = CollectInformationFlowStep._default_utter(slot_name)

        if step_data.get("utter") == default_utter:
            step_data.pop("utter", None)
        if step_data.get("ask_before_filling") is DEFAULT_ASK_BEFORE_FILLING:
            step_data.pop("ask_before_filling", None)
        if step_data.get("reset_after_flow_ends") is DEFAULT_RESET_AFTER_FLOW_ENDS:
            step_data.pop("reset_after_flow_ends", None)
        if step_data.get("force_slot_filling") is DEFAULT_FORCE_SLOT_FILLING:
            step_data.pop("force_slot_filling", None)
        if not step_data.get("rejections"):
            step_data.pop("rejections", None)

    clean_flow_data(flow_data)

    return flow_data



import json
import os
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional

import rasa.shared.utils.io
import rasa.utils.io
from rasa.llm_fine_tuning.conversations import Conversation
from rasa.shared.utils.yaml import write_yaml

if TYPE_CHECKING:
    from rasa.e2e_test.e2e_test_case import TestSuite
    from rasa.llm_fine_tuning.llm_data_preparation_module import LLMDataExample
    from rasa.llm_fine_tuning.train_test_split_module import DataExampleFormat


class StorageType(Enum):
    FILE = "file"
    # We might want to add other storage systems in the future
    # CLOUD = "cloud"
    # DATABASE = "database"


class StorageStrategy(ABC):
    @abstractmethod
    def write_conversation(
        self, conversation: Conversation, storage_location: Optional[str] = None
    ) -> None:
        pass

    def write_llm_data(
        self, llm_data: List["LLMDataExample"], storage_location: Optional[str]
    ) -> None:
        pass

    def write_formatted_finetuning_data(
        self,
        formatted_data: List["DataExampleFormat"],
        module_storage_location: Optional[str],
        file_name: Optional[str],
    ) -> None:
        pass

    def write_e2e_test_suite_to_yaml_file(
        self,
        e2e_test_suite: "TestSuite",
        module_storage_location: Optional[str],
        file_name: Optional[str],
    ) -> None:
        pass


class FileStorageStrategy(StorageStrategy):
    def __init__(self, output_dir: str) -> None:
        self.output_dir = output_dir

    def write_conversation(
        self, conversation: Conversation, storage_location: Optional[str] = None
    ) -> None:
        file_name = os.path.basename(
            conversation.original_e2e_test_case.file or f"{conversation.name}.yaml"
        )
        file_path = self._get_file_path(storage_location, file_name)
        self._create_output_dir(file_path)

        rasa.utils.io.write_yaml(conversation.as_dict(), file_path)

    def _get_file_path(
        self, storage_location: Optional[str], file_name: Optional[str] = None
    ) -> Path:
        if storage_location:
            output = f"{self.output_dir}/{storage_location}"
        else:
            output = self.output_dir

        if file_name:
            return Path(f"{output}/{file_name}")
        return Path(output)

    @staticmethod
    def _create_output_dir(file_path: Path) -> None:
        if (
            str(file_path).endswith(".jsonl")
            or str(file_path).endswith(".yaml")
            or str(file_path).endswith(".yml")
        ):
            file_path.parent.mkdir(exist_ok=True, parents=True)
        else:
            file_path.mkdir(exist_ok=True, parents=True)

    def write_llm_data(
        self, llm_data: List["LLMDataExample"], storage_location: Optional[str]
    ) -> None:
        file_path = self._get_file_path(storage_location)
        self._create_output_dir(file_path)

        with open(str(file_path), "w", encoding="utf-8") as outfile:
            for example in llm_data:
                json.dump(example.as_dict(), outfile, ensure_ascii=False)
                outfile.write("\n")

    def write_formatted_finetuning_data(
        self,
        formatted_data: List["DataExampleFormat"],
        module_storage_location: Optional[str],
        file_name: Optional[str],
    ) -> None:
        file_path = self._get_file_path(module_storage_location, file_name)
        self._create_output_dir(file_path)

        with open(str(file_path), "w", encoding="utf-8") as file:
            for example in formatted_data:
                json.dump(example.as_dict(), file, ensure_ascii=False)
                file.write("\n")

    def write_e2e_test_suite_to_yaml_file(
        self,
        e2e_test_suite: "TestSuite",
        module_storage_location: Optional[str],
        file_name: Optional[str],
    ) -> None:
        """Write the e2e TestSuite data to a YAML file.

        Args:
            e2e_test_suite: The TestSuite object to write.
            module_storage_location: The location where the file should be stored.
            file_name: The path to the file where the data should be written.
        """
        file_path = self._get_file_path(module_storage_location, file_name)
        self._create_output_dir(file_path)
        write_yaml(e2e_test_suite.as_dict(), str(file_path))


class StorageContext:
    def __init__(self, strategy: StorageStrategy) -> None:
        self.strategy = strategy

    def write_conversation(
        self, conversation: Conversation, sub_dir: Optional[str] = None
    ) -> None:
        self.strategy.write_conversation(conversation, sub_dir)

    def write_conversations(
        self, conversations: List[Conversation], sub_dir: Optional[str] = None
    ) -> None:
        for conversation in conversations:
            self.write_conversation(conversation, sub_dir)

    def write_llm_data(
        self, llm_data: List["LLMDataExample"], file_path: Optional[str] = None
    ) -> None:
        self.strategy.write_llm_data(llm_data, file_path)

    def write_formatted_finetuning_data(
        self,
        formatted_data: List["DataExampleFormat"],
        module_storage_location: Optional[str] = None,
        file_name: Optional[str] = None,
    ) -> None:
        self.strategy.write_formatted_finetuning_data(
            formatted_data, module_storage_location, file_name
        )

    def write_e2e_test_suite_to_yaml_file(
        self,
        e2e_test_suite: "TestSuite",
        module_storage_location: Optional[str] = None,
        file_name: Optional[str] = None,
    ) -> None:
        self.strategy.write_e2e_test_suite_to_yaml_file(
            e2e_test_suite, module_storage_location, file_name
        )



from typing import Any, Dict, Optional

import boto3
import structlog
from botocore.exceptions import BotoCoreError, ClientError

from rasa.shared.constants import (
    API_BASE_CONFIG_KEY,
    API_VERSION_CONFIG_KEY,
    AWS_ACCESS_KEY_ID_CONFIG_KEY,
    AWS_BEDROCK_PROVIDER,
    AWS_REGION_NAME_CONFIG_KEY,
    AWS_SAGEMAKER_CHAT_PROVIDER,
    AWS_SAGEMAKER_PROVIDER,
    AWS_SECRET_ACCESS_KEY_CONFIG_KEY,
    AWS_SESSION_TOKEN_CONFIG_KEY,
    AZURE_API_BASE_ENV_VAR,
    AZURE_API_VERSION_ENV_VAR,
    DEPLOYMENT_CONFIG_KEY,
)
from rasa.shared.exceptions import ProviderClientValidationError
from rasa.shared.utils.io import resolve_environment_variables

structlogger = structlog.get_logger()


def validate_aws_setup_for_litellm_clients(
    litellm_model_name: str, litellm_call_kwargs: Dict, source_log: str, provider: str
) -> None:
    """Validates the AWS setup for LiteLLM clients to ensure credentials are set.

    Args:
        litellm_model_name (str): The name of the LiteLLM model being validated.
        litellm_call_kwargs (dict): Additional keyword arguments passed to the client,
            which may include configuration values for AWS credentials.
        source_log (str): The source log identifier for structured logging.
        provider (str): The provider for which the validation is being performed.

    Raises:
        ProviderClientValidationError: If any required AWS environment variable
            or corresponding configuration key is missing.
    """
    # expand environment variables if referenced in the config
    resolved_litellm_call_kwargs: Dict = resolve_environment_variables(
        litellm_call_kwargs
    )  # type: ignore[assignment]

    # boto3 only accepts bedrock and sagemaker as valid clients
    # therefore we need to convert the provider name if it is defined
    # as sagemaker_chat
    provider = (
        AWS_SAGEMAKER_PROVIDER if provider == AWS_SAGEMAKER_CHAT_PROVIDER else provider
    )

    # if the AWS credentials are defined in the endpoints yaml model config,
    # either as referenced secret env vars or direct values, we need to pass them
    # to the boto3 client to ensure that the client can connect to the AWS service.
    additional_kwargs: Dict[str, Any] = {}
    if AWS_ACCESS_KEY_ID_CONFIG_KEY in resolved_litellm_call_kwargs:
        additional_kwargs[AWS_ACCESS_KEY_ID_CONFIG_KEY] = resolved_litellm_call_kwargs[
            AWS_ACCESS_KEY_ID_CONFIG_KEY
        ]
    if AWS_SECRET_ACCESS_KEY_CONFIG_KEY in resolved_litellm_call_kwargs:
        additional_kwargs[AWS_SECRET_ACCESS_KEY_CONFIG_KEY] = (
            resolved_litellm_call_kwargs[AWS_SECRET_ACCESS_KEY_CONFIG_KEY]
        )
    if AWS_SESSION_TOKEN_CONFIG_KEY in resolved_litellm_call_kwargs:
        additional_kwargs[AWS_SESSION_TOKEN_CONFIG_KEY] = resolved_litellm_call_kwargs[
            AWS_SESSION_TOKEN_CONFIG_KEY
        ]
    if AWS_REGION_NAME_CONFIG_KEY in resolved_litellm_call_kwargs:
        additional_kwargs["region_name"] = resolved_litellm_call_kwargs[
            AWS_REGION_NAME_CONFIG_KEY
        ]

    try:
        # We are using the boto3 client because it can discover the AWS credentials
        # from the environment variables, credentials file, or IAM roles.
        # This is necessary to ensure that the client can connect to the AWS service.
        aws_client = boto3.client(provider, **additional_kwargs)

        # Using different method calls available to different AWS clients
        # to test the connection
        if provider == AWS_SAGEMAKER_PROVIDER:
            aws_client.list_models()
        elif provider == AWS_BEDROCK_PROVIDER:
            aws_client.get_model_invocation_logging_configuration()

    except (ClientError, BotoCoreError) as exc:
        event_info = (
            f"Failed to validate AWS setup for LiteLLM clients: {exc}. "
            f"Ensure that you are using one of the available authentication methods:"
            f"credentials file, environment variables, or IAM roles. "
            f"Also, ensure that the AWS region is set correctly. "
        )
        structlogger.error(
            f"{source_log}.validate_aws_credentials_for_litellm_clients",
            event_info=event_info,
            exception=str(exc),
            model_name=litellm_model_name,
        )
        raise ProviderClientValidationError(event_info)


def validate_azure_client_setup(
    api_base: Optional[str],
    api_version: Optional[str],
    deployment: Optional[str],
) -> None:
    """Validates the Azure setup for LiteLLM Router clients to ensure
     that all required configuration parameters are set.
    Raises:
        ProviderClientValidationError: If any required Azure configurations
            is missing.
    """

    def generate_event_info_for_missing_setting(
        setting: str,
        setting_env_var: Optional[str] = None,
        setting_config_key: Optional[str] = None,
    ) -> str:
        """Generate a part of the message with instructions on what to set
        for the missing client setting.
        """
        info = "Set {setting} with {options}. "
        options = ""
        if setting_env_var is not None:
            options += f"environment variable '{setting_env_var}'"
        if setting_config_key is not None and setting_env_var is not None:
            options += " or "
        if setting_config_key is not None:
            options += f"config key '{setting_config_key}'"

        return info.format(setting=setting, options=options)

    # All required settings for Azure OpenAI client
    settings: Dict[str, Dict[str, Any]] = {
        "API Base": {
            "current_value": api_base,
            "env_var": AZURE_API_BASE_ENV_VAR,
            "config_key": API_BASE_CONFIG_KEY,
        },
        "API Version": {
            "current_value": api_version,
            "env_var": AZURE_API_VERSION_ENV_VAR,
            "config_key": API_VERSION_CONFIG_KEY,
        },
        "Deployment Name": {
            "current_value": deployment,
            "env_var": None,
            "config_key": DEPLOYMENT_CONFIG_KEY,
        },
    }

    missing_settings = [
        setting_name
        for setting_name, setting_info in settings.items()
        if setting_info["current_value"] is None
    ]

    if missing_settings:
        event_info = f"Client settings not set: {', '.join(missing_settings)}. "

        for missing_setting in missing_settings:
            if settings[missing_setting]["current_value"] is not None:
                continue
            event_info += generate_event_info_for_missing_setting(
                missing_setting,
                settings[missing_setting]["env_var"],
                settings[missing_setting]["config_key"],
            )

        structlogger.error(
            "azure_openai_llm_client.not_configured",
            event_info=event_info,
            missing_settings=missing_settings,
        )
        raise ProviderClientValidationError(event_info)



from __future__ import annotations

import contextlib
import itertools
import json
import os
from datetime import datetime
from time import sleep
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    Iterable,
    Iterator,
    Optional,
    Text,
    Union,
)

import sqlalchemy as sa
import structlog

import rasa.shared
from rasa.core.brokers.broker import EventBroker
from rasa.core.constants import (
    POSTGRESQL_MAX_OVERFLOW,
    POSTGRESQL_POOL_SIZE,
    POSTGRESQL_SCHEMA,
)
from rasa.core.tracker_stores.tracker_store import (
    SerializedTrackerAsText,
    TrackerStore,
    validate_port,
)
from rasa.shared.core.domain import Domain
from rasa.shared.core.events import SessionStarted
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.exceptions import RasaException
from rasa.shared.nlu.constants import INTENT_NAME_KEY

if TYPE_CHECKING:
    from sqlalchemy import Sequence
    from sqlalchemy.engine.base import Engine
    from sqlalchemy.engine.url import URL
    from sqlalchemy.orm import Query, Session


structlogger = structlog.get_logger(__name__)

# default values of PostgreSQL pool size and max overflow
POSTGRESQL_DEFAULT_MAX_OVERFLOW = 100
POSTGRESQL_DEFAULT_POOL_SIZE = 50


def _create_sequence(table_name: Text) -> "Sequence":
    """Creates a sequence object for a specific table name.

    If using Oracle you will need to create a sequence in your database,
    as described here: https://rasa.com/docs/rasa-pro/production/tracker-stores#sqltrackerstore
    Args:
        table_name: The name of the table, which gets a Sequence assigned

    Returns: A `Sequence` object
    """
    from sqlalchemy.orm import declarative_base

    sequence_name = f"{table_name}_seq"
    Base = declarative_base()
    return sa.Sequence(sequence_name, metadata=Base.metadata, optional=True)


def is_postgresql_url(url: Union[Text, "URL"]) -> bool:
    """Determine whether `url` configures a PostgreSQL connection.

    Args:
        url: SQL connection URL.

    Returns:
        `True` if `url` is a PostgreSQL connection URL.
    """
    if isinstance(url, str):
        return "postgresql" in url

    return url.drivername == "postgresql"


def create_engine_kwargs(url: Union[Text, "URL"]) -> Dict[Text, Any]:
    """Get `sqlalchemy.create_engine()` kwargs.

    Args:
        url: SQL connection URL.

    Returns:
        kwargs to be passed into `sqlalchemy.create_engine()`.
    """
    if not is_postgresql_url(url):
        return {}

    kwargs: Dict[Text, Any] = {}

    schema_name = os.environ.get(POSTGRESQL_SCHEMA)

    if schema_name:
        structlogger.debug(
            "postgresql_tracker_store.schema_name",
            event_inf=f"Using PostgreSQL schema '{schema_name}'.",
        )
        kwargs["connect_args"] = {"options": f"-csearch_path={schema_name}"}

    # pool_size and max_overflow can be set to control the number of
    # connections that are kept in the connection pool. Not available
    # for SQLite, and only  tested for PostgreSQL. See
    # https://docs.sqlalchemy.org/en/13/core/pooling.html#sqlalchemy.pool.QueuePool
    kwargs["pool_size"] = int(
        os.environ.get(POSTGRESQL_POOL_SIZE, POSTGRESQL_DEFAULT_POOL_SIZE)
    )
    kwargs["max_overflow"] = int(
        os.environ.get(POSTGRESQL_MAX_OVERFLOW, POSTGRESQL_DEFAULT_MAX_OVERFLOW)
    )

    return kwargs


def ensure_schema_exists(session: "Session") -> None:
    """Ensure that the requested PostgreSQL schema exists in the database.

    Args:
        session: Session used to inspect the database.

    Raises:
        `ValueError` if the requested schema does not exist.
        RasaException if no engine can be obtained from session.
    """
    schema_name = os.environ.get(POSTGRESQL_SCHEMA)

    if not schema_name:
        return

    engine = session.get_bind()

    if not isinstance(engine, sa.engine.base.Engine):
        # The "bind" is usually an instance of Engine, except in the case
        # where the session has been explicitly bound directly to a connection.
        raise RasaException("Cannot ensure schema exists as no engine exists.")

    if is_postgresql_url(engine.url):
        query = sa.exists(
            sa.select(sa.text("schema_name"))
            .select_from(sa.text("information_schema.schemata"))
            .where(sa.text(f"schema_name = '{schema_name}'"))
        )
        if not session.query(query).scalar():
            raise ValueError(schema_name)


class SQLTrackerStore(TrackerStore, SerializedTrackerAsText):
    """Store which can save and retrieve trackers from an SQL database."""

    from sqlalchemy.orm import DeclarativeBase

    class Base(DeclarativeBase):
        """Base class for all tracker store tables."""

        pass

    class SQLEvent(Base):
        """Represents an event in the SQL Tracker Store."""

        __tablename__ = "events"

        # `create_sequence` is needed to create a sequence for databases that
        # don't autoincrement Integer primary keys (e.g. Oracle)
        id = sa.Column(sa.Integer, _create_sequence(__tablename__), primary_key=True)
        sender_id = sa.Column(sa.String(255), nullable=False, index=True)
        type_name = sa.Column(sa.String(255), nullable=False)
        timestamp = sa.Column(sa.Float)
        intent_name = sa.Column(sa.String(255))
        action_name = sa.Column(sa.String(255))
        data = sa.Column(sa.Text)

    def __init__(
        self,
        domain: Optional[Domain] = None,
        dialect: Text = "sqlite",
        host: Optional[Text] = None,
        port: Optional[int] = None,
        db: Text = "rasa.db",
        username: Optional[Text] = None,
        password: Optional[Text] = None,
        event_broker: Optional[EventBroker] = None,
        login_db: Optional[Text] = None,
        query: Optional[Dict] = None,
        **kwargs: Dict[Text, Any],
    ) -> None:
        import sqlalchemy.exc

        port = validate_port(port)

        engine_url = self.get_db_url(
            dialect, host, port, db, username, password, login_db, query
        )

        self.engine = sa.create_engine(engine_url, **create_engine_kwargs(engine_url))

        structlogger.debug(
            "sql_tracker_store.connect_to_sql_database",
            event_info=f"Attempting to connect to database via '{self.engine.url!r}'.",
        )

        # Database might take a while to come up
        while True:
            try:
                # if `login_db` has been provided, use current channel with
                # that database to create working database `db`
                if login_db:
                    self._create_database_and_update_engine(db, engine_url)

                try:
                    self.Base.metadata.create_all(self.engine)
                except (
                    sqlalchemy.exc.OperationalError,
                    sqlalchemy.exc.ProgrammingError,
                ) as e:
                    # Several Rasa services started in parallel may attempt to
                    # create tables at the same time. That is okay so long as
                    # the first services finishes the table creation.
                    structlogger.error(
                        "sql_tracker_store.create_tables_failed",
                        event_info="Could not create tables",
                        exec_info=e,
                    )

                self.sessionmaker = sa.orm.session.sessionmaker(bind=self.engine)
                break
            except (
                sqlalchemy.exc.OperationalError,
                sqlalchemy.exc.IntegrityError,
            ) as error:
                structlogger.warning(
                    "sql_tracker_store.initialisation_error",
                    event_info="Failed to establish a connection to the SQL database. ",
                    exc_info=error,
                )
                sleep(5)

        structlogger.debug(
            "sql_tracker_store.connected_to_sql_database",
            event_info=f"Connection to SQL database '{db}' successful.",
        )

        super().__init__(domain, event_broker, **kwargs)

    @staticmethod
    def get_db_url(
        dialect: Text = "sqlite",
        host: Optional[Text] = None,
        port: Optional[int] = None,
        db: Text = "rasa.db",
        username: Optional[Text] = None,
        password: Optional[Text] = None,
        login_db: Optional[Text] = None,
        query: Optional[Dict] = None,
    ) -> Union[Text, "URL"]:
        """Build an SQLAlchemy `URL` object.

        The URL object represents the parameters needed to connect to an
        SQL database.

        Args:
            dialect: SQL database type.
            host: Database network host.
            port: Database network port.
            db: Database name.
            username: User name to use when connecting to the database.
            password: Password for database user.
            login_db: Alternative database name to which initially connect, and create
                the database specified by `db` (PostgreSQL only).
            query: Dictionary of options to be passed to the dialect and/or the
                DBAPI upon connect.

        Returns:
            URL ready to be used with an SQLAlchemy `Engine` object.
        """
        from urllib import parse

        # Users might specify a url in the host
        if host and "://" in host:
            # assumes this is a complete database host name including
            # e.g. `postgres://...`
            return host
        elif host:
            # add fake scheme to properly parse components
            parsed = parse.urlsplit(f"scheme://{host}")

            # users might include the port in the url
            port = parsed.port or port
            host = parsed.hostname or host

        if not query:
            # query needs to be set in order to create a URL
            query = {}

        return sa.engine.url.URL(
            dialect,
            username,
            password,
            host,
            port,
            database=login_db if login_db else db,
            query=query,
        )

    def _create_database_and_update_engine(self, db: Text, engine_url: "URL") -> None:
        """Creates database `db` and updates engine accordingly."""
        from sqlalchemy import create_engine

        if self.engine.dialect.name != "postgresql":
            rasa.shared.utils.io.raise_warning(
                "The parameter 'login_db' can only be used with a postgres database."
            )
            return

        self._create_database(self.engine, db)
        self.engine.dispose()
        engine_url = sa.engine.url.URL(
            drivername=engine_url.drivername,
            username=engine_url.username,
            password=engine_url.password,
            host=engine_url.host,
            port=engine_url.port,
            database=db,
            query=engine_url.query,
        )
        self.engine = create_engine(engine_url)

    @staticmethod
    def _create_database(engine: "Engine", database_name: Text) -> None:
        """Create database `db` on `engine` if it does not exist."""
        import sqlalchemy.exc

        with engine.connect() as connection:
            connection.execution_options(isolation_level="AUTOCOMMIT")
            matching_rows = connection.execute(
                sa.text(
                    f"SELECT 1 FROM pg_catalog.pg_database "
                    f"WHERE datname = '{database_name}'"
                )
            ).rowcount

            if not matching_rows:
                try:
                    connection.execute(sa.text(f"CREATE DATABASE {database_name}"))
                except (
                    sqlalchemy.exc.ProgrammingError,
                    sqlalchemy.exc.IntegrityError,
                ) as e:
                    structlogger.error(
                        "sql_tracker_store.create_database_failed",
                        event_info=f"Could not create database '{database_name}'",
                        exec_info=e,
                    )

    @contextlib.contextmanager
    def session_scope(self) -> Generator["Session", None, None]:
        """Provide a transactional scope around a series of operations."""
        session = self.sessionmaker()
        try:
            ensure_schema_exists(session)
            yield session
        except ValueError as e:
            rasa.shared.utils.cli.print_error_and_exit(
                f"Requested PostgreSQL schema '{e}' was not found in the database. To "
                f"continue, please create the schema by running 'CREATE DATABASE {e};' "
                f"or unset the '{POSTGRESQL_SCHEMA}' environment variable in order to "
                f"use the default schema. Exiting application."
            )
        finally:
            session.close()

    async def keys(self) -> Iterable[Text]:
        """Returns sender_ids of the SQLTrackerStore."""
        with self.session_scope() as session:
            sender_ids = session.query(self.SQLEvent.sender_id).distinct().all()
            return [sender_id for (sender_id,) in sender_ids]

    async def delete(self, sender_id: Text) -> None:
        """Delete tracker for the given sender_id."""
        if not await self.exists(sender_id):
            structlogger.info(
                "sql_tracker_store.delete.no_tracker_for_sender_id",
                event_info=f"Could not find tracker for conversation ID '{sender_id}'.",
            )
            return None

        with self.session_scope() as session:
            statement = sa.delete(self.SQLEvent).where(
                self.SQLEvent.sender_id == sender_id
            )
            result = session.execute(statement)
            session.commit()

        structlogger.info(
            "sql_tracker_store.delete.deleted_tracker",
            sender_id=sender_id,
            num_rows=result.rowcount,
        )

    async def retrieve(self, sender_id: Text) -> Optional[DialogueStateTracker]:
        """Retrieves tracker for the latest conversation session."""
        return await self._retrieve(sender_id, fetch_events_from_all_sessions=False)

    async def retrieve_full_tracker(
        self, conversation_id: Text
    ) -> Optional[DialogueStateTracker]:
        """Fetching all tracker events across conversation sessions."""
        return await self._retrieve(
            conversation_id, fetch_events_from_all_sessions=True
        )

    async def count_conversations(self, after_timestamp: float = 0.0) -> int:
        """Returns the number of conversations that have occurred after a timestamp.

        By default, this method returns the number of conversations that
        have occurred after the Unix epoch (i.e. timestamp 0).
        """
        with self.session_scope() as session:
            query = (
                session.query(self.SQLEvent.sender_id)
                .distinct()
                .filter(self.SQLEvent.timestamp >= after_timestamp)
            )
            return query.count()

    async def _retrieve(
        self, sender_id: Text, fetch_events_from_all_sessions: bool
    ) -> Optional[DialogueStateTracker]:
        with self.session_scope() as session:
            serialised_events = self._event_query(
                session,
                sender_id,
                fetch_events_from_all_sessions=fetch_events_from_all_sessions,
            ).all()

            events = [json.loads(event.data) for event in serialised_events]

            if self.domain and len(events) > 0:
                structlogger.debug(
                    "sql_tracker_store.recreating_tracker",
                    event_info=f"Recreating tracker from sender id '{sender_id}'",
                )
                return DialogueStateTracker.from_dict(
                    sender_id, events, self.domain.slots
                )
            else:
                structlogger.debug(
                    "sql_tracker_store._retrieve.no_tracker_for_sender_id",
                    event_info=(
                        f"Can't retrieve tracker matching "
                        f"sender id '{sender_id}' from SQL storage. "
                        f"Returning `None` instead.",
                    ),
                )
                return None

    def _event_query(
        self, session: "Session", sender_id: Text, fetch_events_from_all_sessions: bool
    ) -> "Query":
        """Provide the query to retrieve the conversation events for a specific sender.

        The events are ordered by ID to ensure correct sequence of events.
        As `timestamp` is not guaranteed to be unique and low-precision (float), it
        cannot be used to order the events.

        Args:
            session: Current database session.
            sender_id: Sender id whose conversation events should be retrieved.
            fetch_events_from_all_sessions: Whether to fetch events from all
                conversation sessions. If `False`, only fetch events from the
                latest conversation session.

        Returns:
            Query to get the conversation events.
        """
        # Subquery to find the timestamp of the latest `SessionStarted` event
        session_start_sub_query = (
            session.query(sa.func.max(self.SQLEvent.timestamp).label("session_start"))
            .filter(
                self.SQLEvent.sender_id == sender_id,
                self.SQLEvent.type_name == SessionStarted.type_name,
            )
            .subquery()
        )

        event_query = session.query(self.SQLEvent).filter(
            self.SQLEvent.sender_id == sender_id
        )
        if not fetch_events_from_all_sessions:
            event_query = event_query.filter(
                # Find events after the latest `SessionStarted` event or return all
                # events
                sa.or_(
                    self.SQLEvent.timestamp >= session_start_sub_query.c.session_start,
                    session_start_sub_query.c.session_start.is_(None),
                )
            )

        return event_query.order_by(self.SQLEvent.id)

    async def save(self, tracker: DialogueStateTracker) -> None:
        """Update database with events from the current conversation."""
        await self.stream_events(tracker)

        with self.session_scope() as session:
            # only store recent events
            events = self._additional_events(session, tracker)

            for event in events:
                data = event.as_dict()
                intent = (
                    data.get("parse_data", {}).get("intent", {}).get(INTENT_NAME_KEY)
                )
                action = data.get("name")
                timestamp = data.get("timestamp")

                # noinspection PyArgumentList
                session.add(
                    self.SQLEvent(
                        sender_id=tracker.sender_id,
                        type_name=event.type_name,
                        timestamp=timestamp,
                        intent_name=intent,
                        action_name=action,
                        data=json.dumps(data),
                    )
                )
            session.commit()

        structlogger.debug(
            "sql_tracker_store.save_tracker",
            event_info=(
                f"Tracker with sender_id '{tracker.sender_id}' stored to database",
            ),
        )

    def _additional_events(
        self, session: "Session", tracker: DialogueStateTracker
    ) -> Iterator:
        """Return events from the tracker which aren't currently stored."""
        number_of_events_since_last_session = self._event_query(
            session, tracker.sender_id, fetch_events_from_all_sessions=False
        ).count()

        return itertools.islice(
            tracker.events, number_of_events_since_last_session, len(tracker.events)
        )

    async def update(self, tracker_to_keep: DialogueStateTracker) -> None:
        """Overwrite the tracker in the SQL tracker store."""
        with self.session_scope() as session:
            # Delete events whose timestamp are older
            # than the first event of the tracker to keep.
            statement = sa.delete(self.SQLEvent).where(
                self.SQLEvent.sender_id == tracker_to_keep.sender_id,
                self.SQLEvent.timestamp < tracker_to_keep.events[0].timestamp
                if tracker_to_keep.events
                else 0,
            )

            result = session.execute(statement)
            session.commit()

        first_event_timestamp = str(
            datetime.fromtimestamp(tracker_to_keep.events[0].timestamp)
        )

        structlogger.info(
            "sql_tracker_store.update.updated_tracker",
            sender_id=tracker_to_keep.sender_id,
            first_event_timestamp=first_event_timestamp,
            event_info=f"{result.rowcount} rows removed from tracker.",
        )
